"""Arduino UNO GPIO driver (Arduino must be running pin server program).

set_dac available if an MCP4728 is connected.
https://learn.adafruit.com/adafruit-mcp4728-i2c-quad-dac/arduino

Author: Jacob Feder
Date: 10/17/2022
"""
import serial
import logging
import time

logger = logging.getLogger(__name__)

try:
    from pyqtgraph.Qt import QtCore
    QObject = QtCore.QObject
except ModuleNotFoundError:
    logger.info('Not declaring GUI because the required packages are not available')
    QObject = object
    gui = False
else:
    gui = True

class ArduinoGPIO(QObject):
    if gui:
        # for when the pin state is changed
        # args: pin id, state
        pin_state_changed = QtCore.Signal(int, bool)
        # args: dac channel, dac raw value
        dac_val_changed = QtCore.Signal(str, int)
        # args: dac channel, dac voltage
        dac_voltage_changed = QtCore.Signal(str, float)

    def __init__(self, serial_port: str, pins=None, baud=115200, 
                    timeout: float = 3.0):
        """Args:
            serial_port: serial COM port (see pyserial docs)
            npins: number of digital pins
            baud: baudrate
            timeout: read timeout (s)
        """
        super().__init__()

        self.serial_port = serial_port
        self.baud = baud
        self.timeout = timeout

        # available pins
        if pins is not None:
            self.pins = pins
        else:
            # assume arduino UNO
            self.pins = list(range(2, 13))

        # DAC channel mapping
        self.dac_chs = {
            'VA': 0,
            'VB': 1,
            'VC': 2,
            'VD': 3,
        }

    def open(self):
        self.conn = serial.Serial(self.serial_port, baudrate=self.baud, timeout=self.timeout)
        logger.info(f'connected to Arduino @ [{self.serial_port}]')
        # this is required due to a funky issue with the arduino uno
        # (it restarts the arduino when opening the serial port, so we have to wait for it to boot)
        # https://stackoverflow.com/questions/1618141/pyserial-problem-with-arduino-works-with-the-python-shell-but-not-in-a-program/4941880#4941880
        time.sleep(2)

    def close(self):
        self.conn.close()

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *args):
        self.close()

    def _send_cmd(self, cmd, arg1=0, arg2=0, arg3=0, tries=0):
        """Format and send a command to the device.
        
        Args:
            cmd: the command to send, e.g. 'o' for digital output
            arg1: first argument for command
            arg2: second argument for command
            tries: the number of times that the communication has been attempted

        Returns:
            The device response as a string
        """
        msg = cmd.encode('ascii') + bytes([arg1, arg2, arg3])
        logger.debug(f'writing packet 0x{msg.hex()} to Arduino [{self.conn}]')
        self.conn.write(msg)
        # response the Arduino sends when it's ready for the next command
        ok_resp = b'rdy\n'
        # response from the Arduino
        resp = b''
        while ok_resp not in resp:
            resp += self.conn.read(1)
            logger.debug(f'Arduino responded with [{resp.decode("ascii", "ignore").encode("unicode_escape")} (ascii)] [{resp.hex()}(hex)]')
            if resp == b'':
                raise RuntimeError('Error in communication with Arduino program. No Response.')
        if b'error' in resp:
            if tries < 5:
                # shift one byte so that hopefully it reads correctly next time
                self.conn.write(bytes([0]))
                self._send_cmd(cmd, arg1, arg2, arg3, tries=tries + 1)
            else:
                # if it's failed too many times, give up
                raise RuntimeError(f'Error in communication with Arduino program. Arduino program error:\n{resp}')
        return resp

    def set_pin(self, pin, state, emit=True):
        """Set the output state of a GPIO pin.

        Args:
            pin: pin number
            state: True for high, False for low
            emit: True to emit a PyQt signal when the function finishes

        Raises:
            ValueError: invalid pin
        """
        if pin not in self.pins:
            raise ValueError(f'Pin is invalid. Available pins:\n{self.pins}')
        if state:
            self._send_cmd('o', pin, 1)
            logger.debug(f'Set Arduino@{self.conn} pin [{pin}] HIGH')
        else:
            self._send_cmd('o', pin, 0)
            logger.debug(f'Set Arduino@{self.conn} pin [{pin}] LOW')
        if gui and emit:
            self.pin_state_changed.emit(pin, state)

    def set_dac(self, ch, val, emit=True):
        """Set the output voltage of the connected MCP4728 quad DAC.

        Args:
            ch: DAC channel string ('VA', 'VB', 'VC', or 'VD')
            val: 12-bit DAC output raw value (0-4095) 
            emit: True to emit a PyQt signal when the function finishes

        Raises:
            ValueError: invalid pin
        """
        if ch not in self.dac_chs:
            raise ValueError(f'DAC channel is invalid. Available channels:\n{self.dac_chs.keys()}')
        if val < 0 or val > 4095 or not isinstance(val, int):
            raise ValueError('val must be an int in the range 0 < val < 4095')

        self._send_cmd('d', self.dac_chs[ch], val & 0xFF, (val & 0xFF00) >> 8)
        logger.debug(f'Set Arduino@{self.conn} dac ch [{ch}] to [{val}]')

        if gui and emit:
            self.dac_val_changed.emit(ch, val)

    def set_voltage(self, ch, volt, emit=True):
        """Set DAC output voltage
        
        Args:
            ch: DAC channel string ('VA', 'VB', 'VC', or 'VD')
            volt: voltage value between 0 and 4.096V
            emit: True to emit a PyQt signal when the function finishes
        """
        val = round(volt / 4.096 * 4095)
        self.set_dac(ch, val, emit=emit)
        if gui and emit:
            self.dac_voltage_changed.emit(ch, volt)

if __name__ == '__main__':
    import time
    logging.basicConfig(level=logging.DEBUG, format='%(asctime)s.%(msecs)03d [%(levelname)8s] %(message)s', datefmt='%m-%d-%Y %H:%M:%S')

    with ArduinoGPIO('/dev/serial/by-id/usb-Arduino__www.arduino.cc__0043_95138323838351118221-if00') as arduino:
        arduino.set_pin(2, True)
        time.sleep(2)
        arduino.set_pin(2, False)
        for i in range(0, 4096, 10):
            arduino.set_dac('VA', i)
            time.sleep(0.01)



"""
GUI for streaming XIMEA camera video.

Copyright (c) 2022, Jacob Feder
All rights reserved.
"""
from functools import partial
import logging
from threading import Lock

import numpy as np
from pyqtgraph.Qt import QtGui
from pyqtgraph.Qt import QtCore
from pyqtgraph.Qt import QtWidgets
from pyqtgraph import SpinBox

from nspyre_drivers.ximea.ximea import XIMEA

logger = logging.getLogger(__name__)

class VideoStreamer(QtCore.QObject):
    """Worker object to stream data from the camera in a thread."""
    new_frame = QtCore.Signal(np.ndarray)

    def __init__(self, cam):
        super().__init__()
        self.cam = cam
        self.acquiring = False
        self.lock = Lock()
        self.sem = QtCore.QSemaphore(n=1)

    def acquire(self):
        """Continuously acquire images."""
        with self.lock:
            if self.acquiring:
                logger.warning('Cannot start acquisition because camera is already acquiring.')
                return
            self.acquiring = True
        while self.acquiring:
            # don't get the next frame until the previous frame has been displayed
            # in order to prevent overwhelming the GUI
            self.sem.acquire()
            frame = self.cam.get_image()
            self.new_frame.emit(frame)


class XimeaCameraWidget(QtWidgets.QWidget):
    """Qt widget for displaying the camera feed."""
    def __init__(self, camera_port=None):
        """
        Args:
            camera_port: the serial number of the camera, or None to pick the first available
        """
        super().__init__()

        # create camera driver
        self.cam = XIMEA(camera_port)
        self.cam.__enter__()
        self.cam.start_acquisition()

        # top level layout
        layout = QtWidgets.QVBoxLayout()

        # get an image to figure out the dimension
        img = self.cam.get_image()
        self.disply_width = img.shape[0]
        self.display_height = img.shape[1]

        # create the label that holds the image
        self.image_label = QtWidgets.QLabel()
        self.image_label.resize(self.disply_width, self.display_height)
        layout.addWidget(self.image_label)

        # worker object to read the camera data
        self.stream_worker = VideoStreamer(self.cam)
        # proper Qt thread handling
        # https://mayaposch.wordpress.com/2011/11/01/how-to-really-truly-use-qthreads-the-full-explanation/
        self.thread = QtCore.QThread()
        self.stream_worker.moveToThread(self.thread)
        # stop the streamer when this object is destroyed
        self.destroyed.connect(partial(self.stop_streamer))
        self.destroyed.connect(self.thread.quit)
        self.thread.finished.connect(partial(self.exit))
        # run update_image to update the QLabel whenever a new image is available from the camera
        self.stream_worker.new_frame.connect(self.update_image)
        # start the thread
        self.thread.start()

        # acuisition controls
        controls_layout = QtWidgets.QHBoxLayout()

        # button to turn the acquisition on
        self.start_button = QtWidgets.QPushButton('Start')
        self.start_button.clicked.connect(self.stream_worker.acquire)
        controls_layout.addWidget(self.start_button)

        # button to turn the acquisition off
        self.stop_button = QtWidgets.QPushButton('Stop')
        self.stop_button.clicked.connect(self.stop_streamer)
        controls_layout.addWidget(self.stop_button)

        # exposure time spinbox
        controls_layout.addWidget(QtWidgets.QLabel('Exposure time'))
        self.exposure_spinbox = SpinBox(value=100e-3, suffix='s', siPrefix=True, bounds=(0, 1000), dec=True, minStep=1e-6)
        self.exposure_spinbox.sigValueChanged.connect(lambda spinbox: self.cam.driver.set_exposure(spinbox.value()*1e6))
        self.exposure_spinbox.setValue(100e-3)
        controls_layout.addWidget(self.exposure_spinbox)

        # gain spinbox
        controls_layout.addWidget(QtWidgets.QLabel('Gain'))
        self.gain_spinbox = SpinBox(value=1, bounds=(0, 2e16), dec=True, minStep=1e-3)
        self.gain_spinbox.sigValueChanged.connect(lambda spinbox: self.cam.driver.set_gain(spinbox.value()))
        self.gain_spinbox.setValue(1)
        controls_layout.addWidget(self.gain_spinbox)

        layout.addLayout(controls_layout)
        self.setLayout(layout)

    def stop_streamer(self):
        self.stream_worker.acquiring = False

    def exit(self):
        self.cam.stop_acquisition()
        self.cam.__exit__()

    def update_image(self, frame):
        """Updates the image_label with a new opencv image"""
        # convert from numpy array to qt pixmap
        qt_img = self.convert_np_qt(frame)
        self.image_label.setPixmap(qt_img)
        self.stream_worker.sem.release()

    def convert_np_qt(self, img):
        """Convert from an greyscale numpy array image to QPixmap"""
        # make a copy to prevent the data from being overwritten
        self.img = img.copy()
        width = img.shape[0]
        height = img.shape[1]
        qt_image = QtGui.QImage(img, width, height, QtGui.QImage.Format.Format_Grayscale8)
        # scale to the window size
        scaled_qt_image = qt_image.scaled(self.image_label.frameGeometry().width(),
                                        self.image_label.frameGeometry().height(),
                                        QtCore.Qt.AspectRatioMode.KeepAspectRatio)
        return QtGui.QPixmap.fromImage(scaled_qt_image)



"""
Qt GUI for Ocean Optics spectrometer HR2000+ES.

Copyright (c) 2022, Benjamin Soloway
All rights reserved.
"""
import logging

from pyqtgraph.Qt import QtWidgets
from pyqtgraph import SpinBox
from nspyre import LinePlotWidget
from rpyc.utils.classic import obtain

logger = logging.getLogger(__name__)

class OceanWidget(QtWidgets.QWidget):
    """Qt widget for controlling Ocean Optics HR2000ES spectrometers."""

    def __init__(self, spectrometer_driver):
        """
        Args:
            spectrometer_driver: The HR2000ES driver.
        """
        super().__init__()

        self.spec = spectrometer_driver

        # top level layout
        layout = QtWidgets.QGridLayout()
        layout_row = 0

        # button to acquire data
        self.plot = HR2000ES_LinePlotWidget()
        layout.addWidget(self.plot, layout_row, 0)
        layout_row+=1

        self.acquire_button = QtWidgets.QPushButton('Acquire')
        self.acquire_button.clicked.connect(self.acquire_and_plot)
        layout.addWidget(self.acquire_button, layout_row, 0)
        layout_row += 1

        # integration time spinbox
        layout.addWidget(QtWidgets.QLabel('Integration time'), layout_row, 0)
        layout_row += 1
        self.int_time_spinbox = SpinBox(value=0.1, suffix='s', siPrefix=True, bounds=(0, 4.147e6), dec=True, minStep=0.1e-3)
        self.int_time_spinbox.sigValueChanged.connect(lambda spinbox: self.spec.set_int_time_sec(spinbox.value()))
        self.int_time_spinbox.setValue(value=0.1)
        layout.addWidget(self.int_time_spinbox, layout_row, 0)
        layout_row += 1

        # time acquiring label
        self.time_acq_label = QtWidgets.QLabel('')
        layout.addWidget(self.time_acq_label, layout_row, 1)
        layout_row += 1
        # get time acquiring
        self.get_time_acq_button = QtWidgets.QPushButton('Get elapsed acquisition duration')
        self.get_time_acq_button.clicked.connect(self.update_acquisition_duration)
        layout.addWidget(self.get_time_acq_button, layout_row, 0)
        layout_row += 1

        # trigger mode spinbox
        layout.addWidget(QtWidgets.QLabel('Trigger Mode'), layout_row, 0)
        self.port_dropdown = QtWidgets.QComboBox()
        self.port_dropdown.addItem('Normal ("Free Running")') # index 0
        self.port_dropdown.addItem('Software Trigger (not tested)') # index 1
        self.port_dropdown.addItem('External Hardware Level (not tested)')  # index 2
        self.port_dropdown.addItem('External Hardware Edge')  # index 3
        layout_row += 1
        # default to modulation mode
        self.port_dropdown.setCurrentIndex(0)
        self.port_dropdown.currentIndexChanged.connect(self.set_trigger_mode)
        layout.addWidget(self.port_dropdown, layout_row, 0)
        layout_row += 1

        # #trigger mode label
        # self.trigger_label = QLabel('')
        # layout.addWidget(self.trigger_label, layout_row, 1)
        # layout_row += 1
        # # get time acquiring
        # self.get_trig_mode_button = QPushButton('Get trigger mode')
        # self.get_trig_mode_button.clicked.connect(self.update_trig_mode)
        # layout.addWidget(self.get_trig_mode_button, layout_row, 0)
        # layout_row += 1

        ## button to turn the laser on
        #self.on_button = QPushButton('On')
        #self.on_button.clicked.connect(lambda: self.laser.turn_on())
        #self.on_button.clicked.connect(self.update_state)
        #layout.addWidget(self.on_button, layout_row, 1)
        #layout_row += 1
        self.setLayout(layout)

    def acquire_and_plot(self):
        ws, ints = obtain(self.spec.acquire())
        self.plot.set_data('Spectrum', ws, ints)

    def update_time_acquiring(self):
        """Query the spectrometer for the current state then update the state text box."""
        self.time_acq_label.setText(self.spec.get_state()[4:])

    def update_acquisition_duration(self):
        self.time_acq_label.setText(str(self.spec.get_elapsed_time_sec()))

    def set_trigger_mode(self, idx):
        self.spec.set_trigger_mode(idx)

    # def update_trig_mode(self):
    #     """Query the laser for the current state then update the state text box."""
    #     self.trigger_label.setText(self.spec.get_state()[4:])



# -*- coding: utf-8 -*-
"""
Created on Fri Jun  5 15:58:47 2020

@author: Ocean Insight Inc.
"""

from oceandirect.OceanDirectAPI import OceanDirectAPI, OceanDirectError, FeatureID

def printBinary(value):
    bitCount = 32 #32 bits
    numValue = int(value)

    print("bits (%d) = " % numValue, end='')
    for i in range(32, -1, -1):
        print("%d " % ((numValue >> i) & 0x01), end='')

    print(" ")

def acquisitionDelay(device):
    #device.set_acquisition_delay(120)
    #print("acquisitionDelay(device): set acqDelay 120")

    acqDelay    = device.get_acquisition_delay()
    acqDelayInc = device.get_acquisition_delay_increment()
    acqDelayMin = device.get_acquisition_delay_minimum()
    acqDelayMax = device.get_acquisition_delay_maximum()

    print("acquisitionDelay(device): acqDelay     =  %d " % acqDelay)
    print("acquisitionDelay(device): acqDelayInc  =  %d " % acqDelayInc)
    print("acquisitionDelay(device): acqDelayMin  =  %d " % acqDelayMin)
    print("acquisitionDelay(device): acqDelayMax  =  %d " % acqDelayMax)
    print("")

    #400us
    value = 400
    device.set_acquisition_delay(value)
    print("acquisitionDelay(device): set acqDelay =  %d " % value)

    acqDelay = device.get_acquisition_delay()
    print("acquisitionDelay(device): get acqDelay(expected 400)  =  %d " % acqDelay)
    print("")


def gpio(device):
    supported = device.is_feature_id_enabled(FeatureID.GPIO)
    print("gpio(device): GPIO feature supported  = %s" % supported)

    if not supported:
        print("")
        return

    tempCount = device.Advanced.get_gpio_pin_count()
    print("gpio(device): pin count          =  %d " % tempCount)

    try:
        outputVector = device.Advanced.gpio_get_output_enable2()
        printBinary(outputVector)
        print("---------------------------------------------------------------\n\n")
    except OceanDirectError as err:
        [errorCode, errorMsg] = err.get_error_details()
        print("gpio(device): exception / %d = %s" % (errorCode, errorMsg))

    try:
        device.Advanced.gpio_set_output_enable1(0, 1)
        bitOutput1 = device.Advanced.gpio_get_output_enable1(0)
        print("gpio(device): set output vector bits / mask     =  00000000 / 0(True)")
        print("gpio(device): get bit(0) output                 =  0(%s)" % bitOutput1)
        outputVector = device.Advanced.gpio_get_output_enable2()
        print("gpio(device): get output vector expected output =  00000001")
        printBinary(outputVector)
        print("\n")

        device.Advanced.gpio_set_output_enable1(2, 1)
        device.Advanced.gpio_set_output_enable1(3, 1)
        bitOutput1 = device.Advanced.gpio_get_output_enable1(2)
        bitOutput2 = device.Advanced.gpio_get_output_enable1(3)
        print("gpio(device): set output vector bits / mask     =  00000001 / 2(True),3(True)")
        print("gpio(device): get bit(2,3) output               =  2(%s) / 3(%s)" % (bitOutput1, bitOutput2) )
        outputVector = device.Advanced.gpio_get_output_enable2()
        print("gpio(device): get output vector expected output =  00001101")
        printBinary(outputVector)
        print("\n")

        device.Advanced.gpio_set_output_enable1(0, 0)
        device.Advanced.gpio_set_output_enable1(1, 1)
        device.Advanced.gpio_set_output_enable1(3, 0)
        bitOutput1 = device.Advanced.gpio_get_output_enable1(0)
        bitOutput2 = device.Advanced.gpio_get_output_enable1(1)
        bitOutput3 = device.Advanced.gpio_get_output_enable1(3)
        print("gpio(device): set output vector bits / mask     =  00001101 / 0(False),True(1),3(False)")
        print("gpio(device): get bit(0,1,3) output             =  0(%s) / 1(%s) / 3(%s)" % (bitOutput1, bitOutput2, bitOutput3))
        outputVector = device.Advanced.gpio_get_output_enable2()
        print("gpio(device): get output vector expected output =  00000110")
        printBinary(outputVector)
        print("\n")

        device.Advanced.gpio_set_output_enable1(0, 1)
        device.Advanced.gpio_set_output_enable1(3, 1)
        bitOutput1 = device.Advanced.gpio_get_output_enable1(0)
        bitOutput2 = device.Advanced.gpio_get_output_enable1(3)
        print("gpio(device): set output vector bits / mask     =  00000110 / 0(True),3(True)")
        print("gpio(device): get bit(0,3) output               =  0(%s) / 3(%s)" % (bitOutput1, bitOutput2))
        outputVector = device.Advanced.gpio_get_output_enable2()
        print("gpio(device): get output vector expected output =  00001111")
        printBinary(outputVector)
        print("")
        print("---------------------------------------------------------------\n\n")
    except OceanDirectError as err:
        [errorCode, errorMsg] = err.get_error_details()
        print("gpio(device): gpio_output_bits() / %d = %s" % (errorCode, errorMsg))

    try:
        device.Advanced.gpio_set_value1(0, 1)
        device.Advanced.gpio_set_value1(1, 1)
        value1 = device.Advanced.gpio_get_value1(0)
        value2 = device.Advanced.gpio_get_value1(1)
        print("gpio(device): set value vector values / mask   =  00000000 / 0(True), 1(True)")
        print("gpio(device): get bit(0,1) value               =  0(%s) / 1(%s)" % (value1, value2))
        valueVector = device.Advanced.gpio_get_value2()
        print("gpio(device): get value vector expected values =  00000011")
        printBinary(valueVector)
        print("")

        device.Advanced.gpio_set_value1(2, 1);
        device.Advanced.gpio_set_value1(3, 1);
        value1 = device.Advanced.gpio_get_value1(2)
        value2 = device.Advanced.gpio_get_value1(3)
        print("gpio(device): set value vector values / mask   =  00000011 / 2(True),3(True)")
        print("gpio(device): get bit(2,3) value               =  2(%s) / 3(%s)" % (value1, value2))
        valueVector = device.Advanced.gpio_get_value2()
        print("gpio(device): get value vector expected values =  00001111")
        printBinary(valueVector)
        print("")

        device.Advanced.gpio_set_value1(0, 0)
        device.Advanced.gpio_set_value1(2, 0)
        value1 = device.Advanced.gpio_get_value1(0)
        value2 = device.Advanced.gpio_get_value1(2)
        print("gpio(device): set value vector values / mask   =  00001111 / 0(False), 2(False)")
        print("gpio(device): get bit(0,2) value               =  0(%s) / 2(%s)" % (value1, value2))
        valueVector = device.Advanced.gpio_get_value2()
        print("gpio(device): get value vector expected values =  00001010")
        printBinary(valueVector)
        print("")

        device.Advanced.gpio_set_value1(1, 0)
        device.Advanced.gpio_set_value1(3, 0)
        value1 = device.Advanced.gpio_get_value1(1)
        value2 = device.Advanced.gpio_get_value1(3)
        print("gpio(device): set value vector values / mask   =  00001010 / 1(False), 3(False)")
        print("gpio(device): get bit(1,3) value               =  1(%s) / 3(%s)" % (value1, value2))
        valueVector = device.Advanced.gpio_get_value2()
        print("gpio(device): get value vector expected values =  00000000")
        printBinary(valueVector)
        print("\n")

    except OceanDirectError as err:
        [errorCode, errorMsg] = err.get_error_details()
        print("gpio(device): exception / %d = %s" % (errorCode, errorMsg))

    #Output bit masks
    try:
        #15 = 1111
        device.Advanced.gpio_set_output_enable2(15)
        print("gpio(device): set output mask(15)        =  00001111")

        mask = device.Advanced.gpio_get_output_enable2()
        print("gpio(device): expecting output mask (15) =  %d" % mask)
    except OceanDirectError as err:
        [errorCode, errorMsg] = err.get_error_details()
        print("gpio(device): exception / %d = %s" % (errorCode, errorMsg))

    #Value bit masks
    try:
        #12 = 1100
        device.Advanced.gpio_set_value2(12)
        print("")
        print("gpio(device): set value mask(12)        =  00001100")

        mask = device.Advanced.gpio_get_value2()
        print("gpio(device): expecting value mask (12) =  %d" % mask)
    except OceanDirectError as err:
        [errorCode, errorMsg] = err.get_error_details()
        print("gpio(device): exception / %d = %s" % (errorCode, errorMsg))
    print("")

def singleStrobe(device):
    values = [False, True]
    for enable in values:
        device.Advanced.set_single_strobe_enable(enable);
        print("singleStrobe(device): set enable      =  %s " % enable)

        enable2 = device.Advanced.get_single_strobe_enable();
        print("singleStrobe(device): get enable      =  %s " % enable2)
        print("")

    device.Advanced.set_single_strobe_enable(True);
    enable = device.Advanced.get_single_strobe_enable()
    print("singleStrobe(device): enable(True)    =  %s " % enable)


    strobeWidth = 552
    device.Advanced.set_single_strobe_width(strobeWidth)
    print("singleStrobe(device): set strobeWidth =  %d " % strobeWidth)
    strobeWidth = device.Advanced.get_single_strobe_width()
    print("singleStrobe(device): get strobeWidth =  %d " % strobeWidth)
    print("")

    strobeDelay = 322
    device.Advanced.set_single_strobe_delay(strobeDelay)
    print("singleStrobe(device): set strobeDelay =  %d " % strobeDelay)
    strobeDelay = device.Advanced.get_single_strobe_delay()
    print("singleStrobe(device): get strobeDelay =  %d " % strobeDelay)
    print("")

    enable2 = device.Advanced.get_single_strobe_enable();
    print("singleStrobe(device): get enable      =  %s " % enable2)
    print("")

    try:
        delayMin = device.Advanced.get_single_strobe_delay_minimum()
        print("singleStrobe(device): delayMin         =  %d " % delayMin)
    except OceanDirectError as err:
        [errorCode, errorMsg] = err.get_error_details()
        print("singleStrobe(device): get_single_strobe_delay_minimum() %d = %s" % (errorCode, errorMsg))

    try:
        delayMax = device.Advanced.get_single_strobe_delay_maximum()
        print("singleStrobe(device): delayMax         =  %d " % delayMax)
    except OceanDirectError as err:
        [errorCode, errorMsg] = err.get_error_details()
        print("singleStrobe(device): get_single_strobe_delay_maximum() %d = %s" % (errorCode, errorMsg))

    try:
        delayInc = device.Advanced.get_single_strobe_delay_increment()
        print("singleStrobe(device): delayInc         =  %d " % delayInc)
    except OceanDirectError as err:
        [errorCode, errorMsg] = err.get_error_details()
        print("singleStrobe(device): get_single_strobe_delay_increment() %d = %s" % (errorCode, errorMsg))

    try:
        widthMin = device.Advanced.get_single_strobe_width_minimum()
        print("singleStrobe(device): widthMin         =  %d " % widthMin)
    except OceanDirectError as err:
        [errorCode, errorMsg] = err.get_error_details()
        print("singleStrobe(device): get_single_strobe_width_minimum() %d = %s" % (errorCode, errorMsg))

    try:
        widthMax = device.Advanced.get_single_strobe_width_maximum()
        print("singleStrobe(device): widthMax         =  %d " % widthMax)
    except OceanDirectError as err:
        [errorCode, errorMsg] = err.get_error_details()
        print("singleStrobe(device): get_single_strobe_width_maximum() %d = %s" % (errorCode, errorMsg))

    try:
        widthInc = device.Advanced.get_single_strobe_width_increment()
        print("singleStrobe(device): widthInc         =  %d " % widthInc)
    except OceanDirectError as err:
        [errorCode, errorMsg] = err.get_error_details()
        print("singleStrobe(device): get_single_strobe_width_increment() %d = %s" % (errorCode, errorMsg))

    try:
        cycleMax = device.Advanced.get_single_strobe_cycle_maximum()
        print("singleStrobe(device): cycleMax         =  %d " % cycleMax)
    except OceanDirectError as err:
        [errorCode, errorMsg] = err.get_error_details()
        print("singleStrobe(device): get_single_strobe_cycle_maximum() %d = %s" % (errorCode, errorMsg))
    print("")

def continuousStrobe(device):
    periodInc = 0
    try:
        periodInc = device.Advanced.get_continuous_strobe_period_increment()
        print("continuousStrobe(device): period increment =  %d " % periodInc)
    except OceanDirectError as err:
        [errorCode, errorMsg] = err.get_error_details()
        print("continuousStrobe(device): get_continuous_strobe_period_increment() %d = %s" % (errorCode, errorMsg))

    strobePeriod = device.Advanced.get_continuous_strobe_period()
    print("continuousStrobe(device): get strobePeriod =  %d " % strobePeriod)

    strobeEnable = device.Advanced.get_continuous_strobe_enable()
    print("continuousStrobe(device): get strobeEnable =  %s " % strobeEnable)

    values = [False, True]
    for enable in values:
        device.Advanced.set_continuous_strobe_enable(False)
        print("continuousStrobe(device): get strobeEnable =  %s " % enable)
        strobeEnable = device.Advanced.get_continuous_strobe_enable()
        print("continuousStrobe(device): set strobeEnable =  %s " % enable)
        print("")

    try:
        periodMin = device.Advanced.get_continuous_strobe_period_minimum()
        print("continuousStrobe(device): periodMin          =  %d " % periodMin)
    except OceanDirectError as err:
        [errorCode, errorMsg] = err.get_error_details()
        print("continuousStrobe(device): get_continuous_strobe_period_minimum() %d = %s" % (errorCode, errorMsg))

    try:
        periodMax = device.Advanced.get_continuous_strobe_period_maximum()
        print("continuousStrobe(device): periodMax          =  %d " % periodMax)
    except OceanDirectError as err:
        [errorCode, errorMsg] = err.get_error_details()
        print("continuousStrobe(device): get_continuous_strobe_period_maximum() %d = %s" % (errorCode, errorMsg))



    strobePeriodList = [1200, 1505, 800, 453]
    for period in strobePeriodList:
        if (periodInc > 1) and ((period % periodInc) != 0):
            print("continuousStrobe(device): set strobePeriod =  %d  ====> ********* expecting EXCEPTION!" % period)

        try:
            device.Advanced.set_continuous_strobe_period(period)
            print("continuousStrobe(device): set strobePeriod =  %d " % period)
        except OceanDirectError as err:
            [errorCode, errorMsg] = err.get_error_details()
            print("continuousStrobe(device): set_continuous_strobe_period() %d = %s" % (errorCode, errorMsg))

        try:
            period = device.Advanced.get_continuous_strobe_period()
            print("continuousStrobe(device): get strobePeriod =  %d " % period)
        except OceanDirectError as err:
            [errorCode, errorMsg] = err.get_error_details()
            print("continuousStrobe(device): get_continuous_strobe_period() %d = %s" % (errorCode, errorMsg))
        print("")

    try:
        strobeWidth = 216
        device.Advanced.set_continuous_strobe_width(strobeWidth)
        print("continuousStrobe(device): set strobeWidth    =  %d " % strobeWidth)

        strobeWidth = device.Advanced.get_continuous_strobe_width()
        print("continuousStrobe(device): get strobeWidth    =  %d " % strobeWidth)
    except OceanDirectError as err:
        [errorCode, errorMsg] = err.get_error_details()
        print("continuousStrobe(device): %d = %s" % (errorCode, errorMsg))

    print("")

def scanToAverageBoxcar(device, scanToAve, boxcarWidth):
    try:
        value = device.get_scans_to_average()
        print("scanToAverageBoxcar(): cur scans_to_average        =  %d" % value)

        value = device.get_integration_time()
        print("scanToAverageBoxcar(): current integrationTimeUs   =  %d" % value)

        minAveIntTime = device.get_minimum_averaging_integration_time()
        print("scanToAverageBoxcar(): minAverageIntegrationTimeUs =  %d" % minAveIntTime)
        print("")
    except OceanDirectError as err:
        [errorCode, errorMsg] = err.get_error_details()
        print("scanToAverageBoxcar(): set/get / %d = %s" % (errorCode, errorMsg))

    try:
        print("scanToAverageBoxcar(): set_scans_to_average        =  %d" % scanToAve)
        device.set_scans_to_average(scanToAve)
    except OceanDirectError as err:
        [errorCode, errorMsg] = err.get_error_details()
        print("scanToAverageBoxcar(): ERROR with code/scanToAverage, %d = %s ************" % (errorCode, scanToAve))

    try:
        value = device.get_scans_to_average()
        print("scanToAverageBoxcar(): get_scans_to_average        =  %d" % value)

    except OceanDirectError as err:
        [errorCode, errorMsg] = err.get_error_details()
        print("scanToAverageBoxcar(device): set/get / %d = %s" % (errorCode, errorMsg))

    try:
        device.set_boxcar_width(boxcarWidth)
        print("scanToAverageBoxcar(): set_boxcar_width            =  %d" % boxcarWidth)

        value = device.get_boxcar_width()
        print("scanToAverageBoxcar(): get_boxcar_width            =  %d" % value)

    except OceanDirectError as err:
        [errorCode, errorMsg] = err.get_error_details()
        print("scanToAverageBoxcar(): set/get / %d = %s" % (errorCode, errorMsg))
    print("")

def get_spec_formatted(device, sn):
    try:
        #100ms
        device.set_integration_time(100000);

        print("Reading spectra for dev s/n = %s" % sn, flush=True)
        for i in range(10):
            spectra = device.get_formatted_spectrum()
            print("spectra[100]: %d, %d, %d, %d" % (spectra[100], spectra[101], spectra[102], spectra[103]), flush=True)
    except OceanDirectError as e:
        [errorCode, errorMsg] = err.get_error_details()
        print("get_spec_formatted(device): exception / %d = %s" % (errorCode, errorMsg))

def get_spec_raw_with_meta(device, sn):
    try:
        #100ms
        device.set_integration_time(100000);

        print("[START] Reading spectra for dev s/n = %s" % sn, flush=True)
        for i in range(10):
            spectra = []
            timestamp = []
            total_spectra = device.Advanced.get_raw_spectrum_with_metadata(spectra, timestamp, 3)

            print("len(spectra) =  %d" % (total_spectra) )

            #print sample count on each spectra
            for x in range(total_spectra):
                print("spectraWithMetadata: %d ==>  %d, %d, %d, %d" % (timestamp[x], spectra[x][100], spectra[x][101], spectra[x][102], spectra[x][103]))
    except OceanDirectError as e:
        [errorCode, errorMsg] = err.get_error_details()
        print("get_spec_raw_with_meta(device): exception / %d = %s" % (errorCode, errorMsg))


if __name__ == '__main__':

    #
    # To test this file quickly, do the following:
    # 1. Copy this file into <install_folder>\Python (ex: C:\Program Files\Ocean Insight\OceanDirect SDK-1.31.0\Python)
    # 2. Open a window shell and go to that folder
    # 3. Run this command:  python3 .\PythonExampleForOceanDirect.py
    #

    od = OceanDirectAPI()
    device_count = od.find_usb_devices()
    device_ids = od.get_device_ids()

    device_count = len(device_ids)
    (major, minor, point) = od.get_api_version_numbers()

    print("API Version  : %d.%d.%d " % (major, minor, point))
    print("Total Device : %d     \n" % device_count)

    if device_count == 0:
        print("No device found.")
    else:
        for id in device_ids:
            device       = od.open_device(id)
            serialNumber = device.get_serial_number()

            print("First Device : %d       " % id)
            print("Serial Number: %s     \n" % serialNumber)

            acquisitionDelay(device)
            #scanToAverageBoxcar(device, 10, 0)
            get_spec_formatted(device, serialNumber)
            #If devices don't support metadata then you will get an exception.
            #get_spec_raw_with_meta(device, serialNumber)
            #gpio(device)
            #singleStrobe(device)
            #continuousStrobe(device)

            print("Closing device!\n")
            od.close_device(id)

    print("**** exiting program ****")
    




"""Digital Loggers, Inc. web-controlled power strip part number 222.

https://www.digital-loggers.com/222spec.pdf

This is a thin wrapper around the dlipower module that makes the interface a little more convenient.

There is a bug in the dlipower module that causes things to fail after the session times out.
To increase the session timeout follow the instructions under "Can I set the session timeout longer?"
http://www.digital-loggers.com/lpcfaqs.html#MakeSessionLonger

Author: Jacob Feder
Date: 2/11/2022
"""
import dlipower
import time
from datetime import datetime
import logging

logger = logging.getLogger(__name__)
TIMEOUT = 5.0

class DLIPDU():
    def __init__(self, config=None, hostname='192.168.0.100', userid='admin', password='1234', retries=1, timeout=TIMEOUT, **kwargs):
        """
        Args:
            config: dict with integer keys mapping to the desired names of the outlets.
            setup: Write the system config. This only needs to be done once.
            hostname: See dlipower docs.
            userid: See dlipower docs.
            password: See dlipower docs.
        """
        if config:
            self.config = config
        else:
            # these outlet numbers correspond to the ones printed on the housing
            self.config = {
                            1: 'sw1',
                            2: 'sw2',
                            3: 'sw3',
                            4: 'sw4',
                            5: 'sw5',
                            6: 'sw6',
                            7: 'sw7',
                            8: 'sw8',
                        }
        self.hostname = hostname
        self.userid = userid
        self.password = password
        self.retries = retries
        self.timeout = timeout
        self.kwargs = kwargs

    def __getitem__(self, index_key):
        """Return the outlets object for a given integer index or string key (the name of the power outlets)."""
        def get_switch_index():
            if isinstance(index_key, str):
                # invert the keys and values in the config dict
                inv_config = {v: k for k, v in self.config.items()}
                return inv_config[index_key] - 1
            elif isinstance(index_key, int):
                return index_key - 1
            else:
                raise ValueError('index_key should be an integer or string')

        try:
            outlet = self.switch[get_switch_index()]
        except TypeError as err:
            # this can happen if the switch disconnects for some reason
            logger.debug(f'switch [{self}] is disconnected, attempting reconnection...')
            self.switch.login()
            outlet = self.switch[get_switch_index()]

        return outlet

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, *args):
        pass

    def connect(self):
        """Connect to the switch."""
        # TODO connection timeout not working
        self.switch = dlipower.PowerSwitch(hostname=self.hostname, userid=self.userid, password=self.password, retries=self.retries, timeout=self.timeout, **self.kwargs)

        # gather the current outlet names
        current_outlets = set()
        for s in self.switch[:]:
            current_outlets.add(s.description)

        # check if the outlet names have been set, if not, set up the new outlet names
        if set(self.config.values()) != current_outlets:
            for s in self.config:
                self.switch[s - 1].name = self.config[s]

    def set(self, outlet, state, block=True, timeout=TIMEOUT):
        """Set the state of an outlet(s).
        Args:
            outlet: the name (or index) of the outlet, or an iterable of outlet names (or indices)
            state: True to turn the outlet on, otherwise off
            block: if True, wait until the outlet(s) has been toggled before returning
            timeout: time (s) to wait before giving up
        """
        # create an array of outlets to be rebooted
        if isinstance(outlet, str) or isinstance(outlet, int):
            outlets = [outlet]
        else:
            outlets = outlet

        # set the state of the outlets
        for o in outlets:
            if state:
                self.__getitem__(o).on()
                logger.debug(f'Turned outlet [{o}] on.')
            else:
                self.__getitem__(o).off()
                logger.debug(f'Turned outlet [{o}] off.')

        if block:
            # wait until all the outlets have been set, or timeout
            start_time = datetime.now()
            while True:
                # check if the states of all relevant outlets have changed
                complete = True
                for o in outlets:
                    if (state and self.__getitem__(o).state != 'ON') or \
                        (not state and self.__getitem__(o).state != 'OFF'):
                        complete = False
                if complete:
                    break

                # check for timeout
                time_delta = datetime.now() - start_time
                if time_delta.total_seconds() > timeout:
                    raise TimeoutError(f'Timed out waiting to set the state of the outlets [{outlets}]')

    def on(self, outlet, block=True, timeout=TIMEOUT):
        """Turn an outlet(s) on.
        Args:
            outlet: the name of the outlet, or an iterable of outlet names
            block: if True, wait until the outlet(s) has been toggled before returning
            timeout: time (s) to wait before giving up
        """
        self.set(outlet, True, block=block, timeout=timeout)

    def off(self, outlet, block=True, timeout=TIMEOUT):
        """Turn an outlet(s) off.
        Args:
            outlet: the name of the outlet, or an iterable of outlet names
            block: if True, wait until the outlet(s) has been toggled before returning
            timeout: time (s) to wait before giving up
        """
        self.set(outlet, False, block=block, timeout=timeout)

    def reboot(self, outlet, wait_time=1.0, timeout=TIMEOUT):
        """Power cycle an outlet.
        Args:
            outlet: the name of the outlet, or an iterable of outlet names
            wait_time: the time to wait between turning the outlet off and on
            timeout: time to wait for the individual outlets to toggle before giving up (see set())
        """
        # turn the outlet(s) off
        self.set(outlet, False)

        # wait
        time.sleep(wait_time)

        # turn the outlet(s) on
        self.set(outlet, True)

    def reboot_all(self, wait_time=1.0):
        """Power cycle all outlets.
        Args:
            wait_time: the time to wait between turning the outlet off and on
        """
        self.reboot(self.config.values(), wait_time=wait_time)



"""
    lantz.drivers.ni.ni_motion_controller
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Generalized NI DAQ-based motion controller that uses analog outputs to drive
    motion stages.

    Author: Uri Zvi, Jacob Feder, Aidan Jones
    Date: 9/30/2020
"""

# std lib
from collections import OrderedDict
import time

# 3rd party
import numpy as np
import nidaqmx
from nidaqmx.stream_writers import AnalogMultiChannelWriter
from nidaqmx.stream_readers import CounterReader

# lantz
from lantz import Driver
#from lantz.driver import Feat, DictFeat, Action
from lantz.core.feat import Feat, DictFeat
from lantz.core.action import Action
from lantz import Q_

class NIDAQAxis():
    def __init__(self, ao_ch, units, cal, limits=(None, None)):
        """Class representing an individual axis associated with an individual 
        DAQ analog output (ao) channel
        ao_ch: NI DAQ channel string e.g. 'Dev/ao0'
        units: units string of the axis e.g. 'um'
        cal: conversion of axis units to volts e.g. Q_(10.0, 'V/um')
        limits: limits tuple e.g. (Q_(min, 'um'), Q_(max, 'um'))
        """
        self.ch = ao_ch
        self.units = units
        self.limits = limits
        self.cal = cal

    def enforce_units(self, val):
        """Convert an input quantity object or float into the native units of 
        this axis"""
        if not isinstance(val, Q_):
            val = Q_(val, self.units)
        else:
            val = val.to(self.units)
        return val

    def units_to_volts(self, pos):
        """Convert a quantity in the axis units to a voltage (as a float)"""
        return (self.enforce_units(pos) * self.cal).to('V').m

    def generate_smoothed_pts(self, start_pt, stop_pt, steps):
        """Generate a set of points along a smoothed cosine path between the 
        start and stop points, in order to minimize large accelerations"""
        versine_steps = (1.0 - np.cos(np.linspace(0.0, np.pi, steps))) / 2.0
        return np.ones(steps) * start_pt + versine_steps * (stop_pt - start_pt)

class NIDAQMotionController(Driver):
    def __init__(self, ctr_ch, acq_rate, axes, ao_smooth_steps=Q_(5000, '1/V')):
        """Motion controller for an n-dimensional set of NI DAQ analog output
        axes
        ctr_ch: string for DAQ name e.g. 'Dev1'
        axes: dictionary of axis name mapped to NIDAQAxis object

        example initialization:
        axis1 = NIDAQAxis('Dev1/ao0', 'um', Q_(10.0, 'V/um'))
        axis2 = NIDAQAxis('Dev1/ao1', 'um', Q_(10.0, 'V/um'))
        mc = NIDAQMotionController('Dev1', {'x' : axis1, 'y' : axis2})
        or
        mc = NIDAQMotionController('Dev1', x=axis1, y=axis2)
        """
        self.axes = OrderedDict(axes)
        self.position = {}
        for a in self.axes:
            self.position[a] = Q_(0.0, self.axes[a].units)
        self.acq_rate = acq_rate
        self.ctr_ch = ctr_ch
        self.ao_smooth_steps = ao_smooth_steps
        self.current_counter_task = None
        self.counter_tasks = []

    def initialize(self):
        self.ao_motion_task = nidaqmx.Task('NIDAQMotionController_AO')
        for a in self.axes:
            min_max_dict = {}
            # TODO allow limits specified in volts as well
            if self.axes[a].limits[0]:
                min_max_dict['min_val'] = self.axes[a].units_to_volts(self.axes[a].limits[0])
            if self.axes[a].limits[1]:
                min_max_dict['max_val'] = self.axes[a].units_to_volts(self.axes[a].limits[1])
            self.ao_motion_task.ao_channels.add_ao_voltage_chan(self.axes[a].ch, 
                                    name_to_assign_to_channel=a,
                                    **min_max_dict)

    def finalize(self):
        self.ao_motion_task.close()
        for t in self.counter_tasks:
            t.close()

    def new_ctr_task(self, ctr_ch):
        self.current_counter_task = nidaqmx.Task('NIDAQMotionController_CTR_{}'.format(np.random.randint(2**31)))
        self.current_counter_task.ci_channels.add_ci_count_edges_chan(ctr_ch)
        self.counter_tasks.append(self.current_counter_task)
        
    def enforce_point_units(self, point):
        """ 

        e.g.
        (if x.units and y.units = 'um')
        self.enforce_point_units(x=Q_(0.5, 'mm'), y=Q_(1.5, 'um')) ->
            {x: Q_(500.0, 'um'), y: Q_(1.5, 'um')}"""
        new_point = {}
        for axis in point:
            new_point[axis] = self.axes[axis].enforce_units(point[axis])
        return new_point

    def move(self, point):
        """Move to a target location using a smooth s-curve-like interpolation
        point: dictionary containing axis names mapped to target values e.g.
        {'x': Q_(0.5, 'um'), 'y': Q_(1.5, 'um')}
        """
        #import pdb; pdb.set_trace()
        target = self.enforce_point_units(point)
        step_voltages = self.smooth_func(self.position, target)
        #print ('I am moving')
        #print(step_voltages)
        #print(type(step_voltages))
        if step_voltages.size:
            self.ao_motion_task.timing.cfg_samp_clk_timing(
                self.acq_rate.to('Hz').m,
                sample_mode=nidaqmx.constants.AcquisitionType.FINITE,
                samps_per_chan=step_voltages.shape[0]
            )
            print("working on line 135")
            self.ao_motion_task.triggers.start_trigger.disable_start_trig()
            sample_writer_stream = AnalogMultiChannelWriter(self.ao_motion_task.out_stream,
                                                                auto_start=False)
            print("working on line 139")
            # must use array with contiguous memory region because NI uses C arrays under the hood
            ni_sample_buffer = np.ascontiguousarray(step_voltages.transpose(), dtype=float)
            print("buffer made")
            # TODO timeout
            sample_writer_stream.write_many_sample(ni_sample_buffer, timeout=10)
            print("samples written")

            self.ao_motion_task.start()
            time.sleep((1.1*step_voltages.shape[0] / self.acq_rate).to('s').m)
            self.ao_motion_task.stop()
        print("move() almost finishes")
        self.position = target

    def line_scan(self, init_point, final_point, steps, pts_per_step=1):
        """1-axis line scan while acquiring counter data"""
        print(init_point)
        print(final_point)
        init_point = self.enforce_point_units(init_point)
        final_point = self.enforce_point_units(final_point)
        print(init_point)
        self.move(init_point)
        step_voltages = self.linear_func(init_point, final_point, steps)
        step_voltages = np.repeat(step_voltages, pts_per_step + 1, axis=0)
        # configure analog output task
        self.ao_motion_task.timing.cfg_samp_clk_timing(
            self.acq_rate.to('Hz').m,
            sample_mode=nidaqmx.constants.AcquisitionType.FINITE,
            samps_per_chan=step_voltages.shape[0]
        )
        #self.ao_motion_task.triggers.start_trigger.disable_start_trig()
        sample_writer_stream = AnalogMultiChannelWriter(self.ao_motion_task.out_stream,
                                                            auto_start=False)
        # must use array with contiguous memory region because NI uses C arrays under the hood
        ni_ao_sample_buffer = np.ascontiguousarray(step_voltages.transpose(), dtype=np.float)
        # TODO timeout
        sample_writer_stream.write_many_sample(ni_ao_sample_buffer, timeout=60)
        
        # e.g. "Dev1"
        device_name = list(self.axes.items())[0][1].ch.split('/')[0]
        
        # configure counter input task
        self.current_counter_task.timing.cfg_samp_clk_timing(
            self.acq_rate.to('Hz').m,
            source='/{}/ao/SampleClock'.format(device_name),
            sample_mode=nidaqmx.constants.AcquisitionType.FINITE,
            samps_per_chan=step_voltages.shape[0]
        )
        # set the counter input to trigger / start acquisition when the AO starts
        self.current_counter_task.triggers.arm_start_trigger.dig_edge_src = '/{}/ao/StartTrigger'.format(device_name)
        self.current_counter_task.triggers.arm_start_trigger.trig_type = nidaqmx.constants.TriggerType.DIGITAL_EDGE
        # create counter stream object
        sample_reader_stream = CounterReader(self.current_counter_task.in_stream)
        # must use array with contiguous memory region because NI uses C arrays under the hood
        ni_ctr_sample_buffer = np.ascontiguousarray(np.zeros(step_voltages.shape[0]), dtype=np.uint32)
        
        # start the move
        self.current_counter_task.start()
        self.ao_motion_task.start()
        # TODO timeout
        sample_reader_stream.read_many_sample_uint32(ni_ctr_sample_buffer,
                                        number_of_samples_per_channel=nidaqmx.constants.READ_ALL_AVAILABLE,
                                        timeout=60)
        
        # wait for motion to complete
        #time.sleep((1.1*step_voltages.shape[0] / self.acq_rate).to('s').m)
        
        self.current_counter_task.stop()
        self.ao_motion_task.stop()
        scanned = ni_ctr_sample_buffer.reshape((steps, pts_per_step+1))
        averaged = np.diff(scanned).mean(axis=1)
        self.position = final_point
        return averaged*self.acq_rate.to('Hz').m
        
    def linear_func(self, start_pt, stop_pt, steps):
        """Generate a set of linearly spaced points between the start and stop point
        return value is in volts"""
        start_volts = np.array([])
        stop_volts = np.array([])
        for axis in self.axes:
            start_volts = np.append(start_volts, self.axes[axis].units_to_volts(start_pt[axis]))
            stop_volts = np.append(stop_volts, self.axes[axis].units_to_volts(stop_pt[axis]))
        linear_steps = np.linspace(0.0, 1.0, steps)
        # pt0 = [0, 0]
        # pt1 =  [1, 2]
        # np.ones = [1, 1, 1, 1, 1, 1, ...]
                        # [[x, x, x, x, x, x....]
                        # [y, y, y, y, y, y, y...]]
                        
                        # [[0.1, 0.2, ..., 1]
                        # [0.2, 0.4, ..., 2]]
        return np.outer(np.ones(steps), start_volts) + np.outer(linear_steps, stop_volts-start_volts)

    def smooth_func(self, start_pt, stop_pt):
        """Generate a set of smooth sinusoidally spaced points between the 
        start and stop point
        return value is in volts"""
        # convert start_pt and stop_pt dictionaries into arrays that contain
        # the corresponding point values in volts
        #import pdb; pdb.set_trace()
        start_volts = np.array([])
        stop_volts = np.array([])
        for axis in self.axes:
            start_volts = np.append(start_volts, self.axes[axis].units_to_volts(start_pt[axis]))
            stop_volts = np.append(stop_volts, self.axes[axis].units_to_volts(stop_pt[axis]))
        print(start_volts, "are our start voltages")
        print(stop_volts, "are our stop voltages")
        steps = int(np.ceil(max(abs(stop_volts - start_volts)) * self.ao_smooth_steps.to('1/V').m))
        # generate a cosine function from 0->pi
        versine_steps = (1.0 - np.cos(np.linspace(0.0, np.pi, steps))) / 2.0
        return np.outer(np.ones(steps), start_volts) + np.outer(versine_steps, stop_volts - start_volts)



"""
Qt GUI for CLD1010 laser driver.

Copyright (c) 2022, Jacob Feder, Ben
All rights reserved.
"""
import logging

from pyqtgraph.Qt import QtWidgets
from pyqtgraph import SpinBox

logger = logging.getLogger(__name__)

class CLD1010Widget(QtWidgets.QWidget):
    """Qt widget for controlling cld1010 lasers."""

    def __init__(self, laser_driver):
        """
        Args:
            laser_driver: The CLD1010 driver.
        """
        super().__init__()

        self.laser = laser_driver

        # top level layout
        layout = QtWidgets.QGridLayout()
        layout_row = 0

        # button to turn the laser off
        self.off_button = QtWidgets.QPushButton('Off')
        self.off_button.clicked.connect(lambda: self.laser.off())
        self.off_button.clicked.connect(self.update_state)
        layout.addWidget(self.off_button, layout_row, 0)

        # button to turn the laser on
        self.on_button = QtWidgets.QPushButton('On')
        self.on_button.clicked.connect(lambda: self.laser.on())
        self.on_button.clicked.connect(self.update_state)
        layout.addWidget(self.on_button, layout_row, 1)
        layout_row += 1

        # power spinbox
        layout.addWidget(QtWidgets.QLabel('Current'), layout_row, 0)
        self.current_spinbox = SpinBox(value=100, suffix='mA', siPrefix=True, bounds=(0, 110), dec=True, minStep=1)
        self.current_spinbox.sigValueChanged.connect(lambda spinbox: self.laser.set_current_setpoint(spinbox.value()/1000))
        self.current_spinbox.setValue(value=0)
        layout.addWidget(self.current_spinbox, layout_row, 1)
        layout_row += 1

        # # dropdown menu to pick laser mode
        # layout.addWidget(QtWidgets.QLabel('Mode'), layout_row, 0)
        # self.mode_dropdown = QComboBox()
        # self.mode_dropdown.addItem('Digital Modulation') # index 0
        # self.mode_dropdown.addItem('Constant Power') # index 1
        # # default to modulation mode
        # self.mode_dropdown.setCurrentIndex(0)
        # self.mode_dropdown.currentIndexChanged.connect(self.change_mode)
        # layout.addWidget(self.mode_dropdown, layout_row, 1)
        # layout_row += 1

        # state label
        self.state_label = QtWidgets.QLabel('')
        self.update_state()
        layout.addWidget(self.state_label, layout_row, 1)
        # get state button
        self.get_state_button = QtWidgets.QPushButton('Get State')
        self.get_state_button.clicked.connect(self.update_state)
        layout.addWidget(self.get_state_button, layout_row, 0)
        layout_row += 1

        # modulation dropdown
        layout.addWidget(QtWidgets.QLabel('Modulation'), layout_row, 0)
        self.modulation_dropdown = QtWidgets.QComboBox()
        self.modulation_dropdown.addItem('Off') # index 0
        self.modulation_dropdown.addItem('On') # index 1
        # default to Modulation Off
        if self.laser.get_modulation_state() == 'Off':
            self.modulation_dropdown.setCurrentIndex(0)
        if self.laser.get_modulation_state() == 'On':
            self.modulation_dropdown.setCurrentIndex(1)
        self.modulation_dropdown.currentIndexChanged.connect(self.set_mod)
        layout.addWidget(self.modulation_dropdown, layout_row, 1)
        layout_row += 1

        # modulation label
        self.mod_label = QtWidgets.QLabel('')
        self.update_mod()
        layout.addWidget(self.mod_label, layout_row, 1)
        # get state button
        self.get_mod_button = QtWidgets.QPushButton('Get Modulation State')
        self.get_mod_button.clicked.connect(self.update_mod)
        layout.addWidget(self.get_mod_button, layout_row, 0)
        layout_row += 1

        # take up any additional space in the final column with padding
        layout.setColumnStretch(2, 1)
        # take up any additional space in the final row with padding
        layout.setRowStretch(layout_row, 1)

        self.setLayout(layout)

    def update_state(self):
        """Query the laser for the current state then update the state text box."""
        self.state_label.setText(str(self.laser.get_ld_state()))

    def update_mod(self):
        """Query the laser for the current state then update the state text box."""
        self.mod_label.setText(str(self.laser.get_modulation_state()))

    def set_mod(self, idx):
        """Change the agiltron port."""
        if idx == 0:
            self.laser.set_modulation_state('Off')
        elif idx == 1:
            self.laser.set_modulation_state('On')
        else:
            raise ValueError(f'Invalid mode selection [{idx}]')
        self.update_mod()


    # TODO modulation mode
    # def change_mode(self, idx):
    #     """Change the laser output mode."""
    #     # constant power
    #     # digital modulation
    #     if idx == 0:
    #         self.laser.modulation_mode()
    #     elif idx == 1:
    #         self.laser.constant_power()
    #     else:
    #         raise ValueError('Invalid mode selection')
    #     self.update_state()



"""Spectrometer/Monochromator from Acton/Princeton Instruments
Using SP2150i but I think this should work for any of their spectrometers

Author: Benjamin Soloway, Jacob Feder
Date: 12/19/2022
"""
import serial
import logging

logger = logging.getLogger(__name__)

class SP2150i():
    def __init__(self, serial_port: str):
        self.serial_port = serial_port

    def _read(self):
        flagStop = 0
        commStr = ''
        while flagStop == 0:
            commStr += self.conn.read(1).decode()
            if 'ok' in commStr:
                flagStop = 1
        logger.debug(f'received [{commStr}]')
        self.clearSer()
        return commStr

    def _write(self, msg):
        enc_msg = msg.encode() + b'\x0d'
        self.conn.write(enc_msg)
        logger.debug(f'sent command [{enc_msg}]')

    def get_all_gratings(self):
        msg = '?GRATINGS'
        self._write(msg)
        return self._read()

    def set_wavelength(self, wavelength):
        """
        wavelength: wavelength in nm as integer or float

        Goes to destination wavelength to nearest 0.1nm at selected scan rate
        """
        self._write(str(wavelength) + ' goto')
        self._read()

    def set_target_wavelength(self, wavelength):
        """
        wavelength: wavelength in nm as integer or float

        Scans to destination wavelength to nearest 0.1nm at sleected scan rate
        """
        self._write(str(wavelength) + ' nm')
        self._read()

    def set_scan_speed(self, speed):
        """
        wavelength: wavelength in nm as integer or float

        Scans to destination wavelength to nearest 0.1nm at sleected scan rate
        """
        self._write(str(speed) + ' nm/min')
        self._read()

    def get_wavelength(self):
        msg = '?NM'
        self._write(msg)
        ans = self._read() #example:  999.989 nm  ok
        return float(self._clean_up_string(ans[1:8])) #example: 999.989

    def get_scanspeed(self):
        msg = '?NM/MIN'
        self._write(msg)
        ans = self._read()  #example: 100.000 nm/min  ok
        return float(self._clean_up_string(ans[1:8])) #example: 100.000

    def get_turret(self):
        """
        Sends the selected turret number (1, 2, 3, or 4) to the computer or terminal.
        """
        msg = '?TURRET'
        self._write(msg)
        ans = self._read()  #example:  1  ok
        return int(ans[1]) #example: 1

    def get_turret_gratings(self):
        """
        Sends the groove spacing of each grating for each turret to the computer or terminal.
        """
        msg = '?TURRETS'
        self._write(msg)
        return self._read()

    def set_grating(self, grating):
        if grating != 1 and grating != 2:
            raise ValueError(f'Grating {grating} is not available.')
        msg = f'{grating} GRATING'
        self._write(msg)
        return self._read()

    def set_turret(self, turret):
        if turret != 1:
            raise ValueError(f'Turret {turret} is not available.')
        msg = f'{turret} TURRET'
        self._write(msg)
        return self._read()

    def open(self):
        self.conn = serial.Serial(port=self.serial_port, baudrate=9600, parity=serial.PARITY_NONE, bytesize=serial.EIGHTBITS, stopbits=serial.STOPBITS_ONE, timeout=10)

    def close(self):
        self.conn.close()

    def clearSer(self):
        self.conn.reset_input_buffer()
        self.conn.reset_output_buffer()

    def __enter__(self):
        self.open()
        self.clearSer()
        return self

    def __exit__(self, *args):
        self.clearSer()
        self.close()

    def _clean_up_string(self, str):
        return str.replace(' ', '').replace('n', '').replace('m', '')

if __name__ == '__main__':
    with Spec('/dev/serial/by-id/usb-Acton_Research_ACTON_RESEARCH_CONTROLLER-if00-port0') as spec:
        print("getting turret")
        print(spec.get_turret())
        print("setting grating 2 - please give this operation at least 20 seconds")
        spec.set_grating(2)
        print("getting all gratings")
        print(spec.get_all_gratings())
        print("setting grating 1  - please give this operation at least 20 seconds")
        spec.set_grating(1)
        print("getting all gratings")
        print(spec.get_all_gratings())
        #spec.get_turret_gratings()
        print("getting scan speed in nm/min")
        print(spec.get_scanspeed())
        print("setting wavelength to 1000nm")
        spec.set_wavelength(1000)
        print("getting wavelength")
        print(spec.get_wavelength())
        print("setting wavelength to 500nm")
        spec.set_wavelength(500)
        print("getting wavelength")
        print(spec.get_wavelength())



"""PI E-710 piezo controller driver. Written for controller firmware version 6.033.

Page numbers refer to the E-710 user manual.

Author: Jacob Feder
Date: August 26 2021
"""

from serial import Serial
import datetime
import logging
import time

logger = logging.getLogger(__name__)

# distance error from target position where the stage can be considered to have reached the target (um)
DEFAULT_EPSILON = 0.05
# time to wait for the stage to reach a target position
DEFAULT_POS_TIMEOUT = 2.0

class PI710:
    def __init__(self, port, axes, timeout=1.0):
        """
        Args:
            port: serial tty port e.g. 'COM4'
            axes: axis mapping from names to channel #'s
                e.g. {'x': '1', 'y': '2', 'z': '3'}
            timeout: timeout for reading (s)
        """
        self.port = port
        # pyserial port object
        self.sp = None
        self.axes = axes
        self.timeout = timeout
        # servo update period
        self.tservo = 1 / 5e3

    def connect(self, baudrate=115200):
        """Connect to the PI controller

        Args:
            baudrate: serial communication baudrate

        Raises:
            ConnectionError: Couldn't connect to the controller
        """
        # initialize at 9600 baud
        self.sp = Serial(self.port, baudrate=9600, timeout=0)
        self.flush()
        # if the controller is in 9600 baud mode, set the new baud rate
        self.send_raw('0BR' + str(baudrate))
        # disconnect then reconnect at the new baudrate
        self.disconnect()
        time.sleep(0.2)
        self.sp = Serial(self.port, baudrate=baudrate, timeout=0)
        # make sure the send buffer is empty
        self.flush()
        # query the device info to confirm connection
        self.send_raw('GI')
        # read whatever info the controller has sent after a small delay
        time.sleep(0.2)
        resp = self.read_all()
        logging.debug(f'Controller info:\n{resp}')
        if 'Digital Piezo Controller' not in resp:
            raise ConnectionError(f'Controller response "{resp}" expected to contain "Digital Piezo Controller". The controller is either not connected or not responding.')
        # disconnect and reconnect one final time, with the proper baud and timeout
        self.disconnect()
        self.sp = Serial(self.port, baudrate=baudrate, timeout=self.timeout)

    def disconnect(self):
        """Close connection to the controller."""
        if self.sp:
            self.sp.close()
            self.sp = None

    def send_raw(self, cmd):
        """Send a message to the controller (terminator is generated automatically).

        Args:
            cmd: command to send as a string
        """
        # add terminator and convert to ascii
        msg = (cmd + '\n').encode('ASCII')
        logger.debug(f'sent [{msg}]')
        # send the message
        self.sp.write(msg)

    def send(self, cmd):
        """Send a message to the controller and check whether it was received and processed (terminator is generated automatically).

        Args:
            cmd: command to send as a string
        """
        # add terminator and convert to ascii
        self.send_raw(cmd)
        # query the controller status
        self.send_raw('1GI8')
        resp = self.readline()
        # extract the status code
        status_code = -1
        for substr in resp.split():
            if substr.isdigit():
                status_code = int(substr)
        if status_code == -1:
            raise ValueError('Error checking controller status code. The command passed to send() may have had a response from the controller - use send_raw() instead.')

        # if the command was not understood, an error bit will be reported (p. 119)
        if status_code & 32768:
            raise ValueError('Command was not understood by the controller')

    def flush(self):
        """Send a terminator character to make sure any previous message has been written."""
        self.send_raw('')

    def readline(self):
        """Read a single newline terminated message from the controller.

        Returns:
            Response from the controller (with the terminator stripped)
        """
        msg = self.sp.readline().decode('ASCII')
        if msg == '':
            raise TimeoutError
        msg = msg.strip('\n')
        logger.debug(f'received [{msg}]')
        return msg

    def read_all(self):
        """Read all lines from the controller available within the timeout period.

        Returns:
            The response from the controller (with line terminators)
        """
        resp = ''
        while True:
            try:
                resp += self.readline() + '\n'
            except TimeoutError:
                return resp

    def servo_mode(self):
        """Enable closed-loop control mode."""
        for axis in self.axes.values():
            # subsequent DW (data write) commands will apply to this axis, RAM only (p. 99)
            self.send(axis + 'DP0')
            # enable servo control (p. 75)
            self.send('2DW1')
            # turn on servo control (p. 141)
            self.send(axis + 'SL1')
        logger.debug('init done')

    def home(self):
        """Go to position 0 on all axes."""
        # TODO autozero AZ?
        for axis in self.axes.values():
            self.send(axis + 'GH')

    def wait_pos(self, axis, target_pos, epsilon, timeout):
        """Poll the stage position until it is within epsilon of the target position.

        Args:
            axis: axis key e.g. 'x'
            target_pos: target position
            epsilon: function will return when the stage is within epsilon of the target
            timeout: wait timeout (s)

        Raises:
            TimeoutError: timeout occurred waiting to reach target position
        """
        start_time = datetime.datetime.now()
        current_pos = self.pos(axis)
        # wait for the stage to be in position for at least 3 subsequent readings
        for i in range(3):
            while abs(target_pos - current_pos) >= epsilon:
                if timeout and (datetime.datetime.now() - start_time).total_seconds() > timeout:
                    raise TimeoutError
                time.sleep(0.01)
                current_pos = self.pos(axis)

    def pos(self, axis):
        """Return the actual position of the given axis.

        Args:
            axis: axis key e.g. 'x'

        Returns:
            The axis position (um)
        """
        if axis in self.axes:
            self.send_raw(self.axes[axis] + 'TP')
            resp = self.readline()
            try:
                pos = float(resp)
            except ValueError as exc:
                raise ConnectionError(f'Invalid response "{resp}" from the controller when querying the position') from exc
            logger.debug(f'[{axis}] at [{pos}]')
            return pos
        else:
            raise ValueError

    def set_pos(self, axis, pos, blocking=True, epsilon=DEFAULT_EPSILON, timeout=DEFAULT_POS_TIMEOUT):
        """Set the target position of the given axis.

        Args:
            axis: axis key e.g. 'x'
            pos: the target position (um)
            blocking: if True, wait until the stage is within epsilon of the target position before returning
            epsilon: see blocking
            timeout: if blocking is True, max time to block (s)

        Raises:
            TimeoutError: timeout occurred waiting to reach target position
        """
        if axis in self.axes:
            logger.info(f'moving [{axis}] to [{pos}]')
            # TODO check input range
            self.send(self.axes[axis] + 'MA' + str(pos))
        else:
            raise ValueError('axis not found')

        if blocking:
            self.wait_pos(axis, pos, epsilon, timeout)

    def config_line_scan(self, axis, scan_len, npts, time_per_pt):
        """Configure the stage to scan in a line. Use line_scan() to initiate the actual scan.

        Args:
            axis: the axis to scan
            scan_len: the distance of the scan (um)
            npts: the number of points in the scan
            time_per_pt: the time to wait at each point (s)

        Raises:
            ValueError: Invalid arguments.
        """
        if axis not in self.axes:
            raise ValueError('axis not found')

        # TODO check input range
        # the servo loop updates with a period of self.tservo.
        # clocks_per_pt is the number of integer multiples of tservo per point
        clocks_per_pt = round(time_per_pt / self.tservo)
        if not clocks_per_pt:
            raise ValueError(f'time_per_pt [{time_per_pt}] must be greater than the controller update rate [{self.tservo}]')
        logger.debug(f'configure line scan [{clocks_per_pt}] clocks_per_pt')

        # number of points allocated for the ramp up / down portion of the motion
        npts_ramp = 50
        # number of points in controller memory
        controller_npts = npts * clocks_per_pt + 2 * npts_ramp
        if controller_npts > 63488:
            raise ValueError('maximum number of points in controller memory exceeded - reduce the npts*time_per_pt')

        if controller_npts > 16384:
            raise ValueError('maximum number of trigger points in controller memory exceeded - reduce npts')

        # correct for ramp up / down portion of the waveform (see p. 121 a2 equation)
        scan_len = scan_len * (controller_npts - npts_ramp) / (controller_npts - 2 * npts_ramp)

        # see p. 54, 119
        # segment definition mode
        self.send('0PT0')
        # set the total number of points in the segment
        self.send('1PT' + str(controller_npts))
        # set the number of points in the curve
        self.send('1CP' + str(controller_npts))
        # set the center point of the curve
        self.send('1PC' + str(round(controller_npts/2)))
        # set the ramp up / down number of points
        self.send('1PS' + str(npts_ramp))
        # set the point index in the segment definition
        # where the curve generation will begin
        self.send('1PA0')
        # set no offset
        self.send('1FO0')
        # generate the line scan curve and save it into the segment
        self.send('1GL' + str(scan_len))
        # give the controller time to generate the points

        # switch from segment definition mode to waveform definition mode
        self.send('0PT0')
        # define waveform as the entire segment
        self.send('1PT' + str(controller_npts))
        # connect waveform with wave generator (p. 58)
        self.send('1SF1')
        # connect desired axis with wave generator
        self.send(self.axes[axis] + 'CF1')

        # configure trigger settings (p. 59)
        # trigger 1 will be a "pixel" pulse that outputs a short pulse at the beginning of each point
        # trigger 2 will be active high during the entire duration of the constant-velocity portion of the line scan
        # trigger 1 is an active-high short pulse
        self.send('1KT1')
        # trigger 2 is an active-high long pulse
        self.send('2KT2')

        # clear all trigger points
        self.send('0FT0')
        # enable trigger 2 on the first point of the constant-velocity portion of the scan
        self.send(f'{npts_ramp + 1}FT2')
        # enable trigger 2 on all subsequent points until the end of the constant-velocity portion of the scan
        self.send(f'{controller_npts - npts_ramp}FT258')
        for i in range(npts):
            # for each "pixel" point, activate trigger 1 (and trigger 2 so as not to undo our work in previous FT statements)
            self.send(f'{npts_ramp + 1 + i * clocks_per_pt}FT3')

    def line_scan(self, axis, start_pos, blocking=True, epsilon=DEFAULT_EPSILON, timeout=DEFAULT_POS_TIMEOUT):
        """Run a line scan that was previously configured with config_line_scan(). The scan will start at the current position. During the scan, generate a pulse on digital output 1 for each point. After motion, return to the start position.

        Args:
            start_pos: starting position of the scan
            blocking, epsilon, timeout: see set_pos

        Raises:
            TimeoutError: timeout occurred waiting to reach target position
        """

        # move to the starting position
        self.set_pos(axis, start_pos, blocking=True, epsilon=epsilon, timeout=timeout)

        # activate the waveform generator a single time, and automatically return to the original start position
        self.send('1SC0')

        # wait until the stage returns to the starting position
        self.wait_pos(axis, start_pos, epsilon, timeout)

    def __enter__(self, *kwargs):
        self.connect()
        self.servo_mode()
        self.home()
        return self

    def __exit__(self, *kwargs):
        self.disconnect()

if __name__ == '__main__':
    # enable logging to console
    logging.basicConfig(level=logging.DEBUG)

    # initialize driver and connect to the controller
    with PI710('COM4', {'x': '1', 'y': '2', 'z': '3'}) as pi:
        # move each axis back and forth a few times
        # for axis, pos in [('x', 100), ('y', 100), ('z', 10)]:
        #     for i in range(3):
        #         pi.set_pos(axis,  0.0, blocking=False)
        #         time.sleep(0.2)
        #         pi.set_pos(axis,  pos, blocking=False)
        #         time.sleep(0.2)

        # do a few line scans
        pi.config_line_scan('x', 200.0, 100, 1e-3)
        for i in range(100):
            pi.line_scan('x', 50)
            time.sleep(1.0)
        print('done')



"""
Qt GUI for acton monochromator driver.

Copyright (c) 2022, Jacob Feder, Benjamin Soloway
All rights reserved.
"""
import logging

from pyqtgraph.Qt import QtWidgets
from pyqtgraph import SpinBox

logger = logging.getLogger(__name__)

class ActonWidget(QtWidgets.QWidget):
    """Qt widget for controlling cobolt lasers."""
    def __init__(self, mono):
        """
        Args:
            mono: The monochromator driver.
        """
        super().__init__()

        self.mono = mono

        # top level layout
        layout = QtWidgets.QGridLayout()
        layout_row = 0

        # wavelength spinbox
        layout.addWidget(QtWidgets.QLabel('Wavelength'), layout_row, 0)
        self.wl_spinbox = SpinBox(value=self.mono.get_wavelength(), suffix='nm', siPrefix=True, bounds=(0, 1000), dec=True, minStep=0.1e-6)
        def set_wl(spinbox):
            self.mono.set_wavelength(spinbox.value())
        self.wl_spinbox.sigValueChanged.connect(set_wl)
        self.wl_spinbox.setValue(value=0)
        layout.addWidget(self.wl_spinbox, layout_row, 1)
        layout_row += 1

        # dropdown menu to pick laser mode
        layout.addWidget(QtWidgets.QLabel('Grating'), layout_row, 0)
        self.mode_dropdown = QtWidgets.QComboBox()
        self.mode_dropdown.addItem(self.mono.get_all_gratings().split('\n')[1]) # index 0
        self.mode_dropdown.addItem(self.mono.get_all_gratings().split('\n')[2]) # index 1
        # default to modulation mode
        self.mode_dropdown.setCurrentIndex(0)
        self.mode_dropdown.currentIndexChanged.connect(self.change_grating)
        layout.addWidget(self.mode_dropdown, layout_row, 1)
        layout_row += 1

        # take up any additional space in the final column with padding
        layout.setColumnStretch(2, 1)
        # take up any additional space in the final row with padding
        layout.setRowStretch(layout_row, 1)

        self.setLayout(layout)

    def update_wavelength(self):
        """Query the laser for the current state then update the state text box."""
        self.state_label.setText(self.mono.get_wavelength())

    def change_grating(self, idx):
        """Change the laser output mode."""
        # constant power
        # digital modulation
        if idx == 0:
            self.mono.set_grating(1)
        elif idx == 1:
            self.mono.set_grating(2)
        else:
            raise ValueError('Invalid mode selection')



"""
Qt GUI for Cobolt lasers utilizing the python driver:
https://github.com/cobolt-lasers/pycobolt

Copyright (c) 2022, Jacob Feder
All rights reserved.
"""
import logging

from pyqtgraph.Qt import QtWidgets
from pyqtgraph import SpinBox

logger = logging.getLogger(__name__)

class CoboltWidget(QtWidgets.QWidget):
    """Qt widget for controlling Cobolt lasers."""
    def __init__(self, laser_driver):
        """
        Args:
            laser_driver: The pycobolt.CoboltLaser driver.
        """
        super().__init__()

        self.laser = laser_driver

        # top level layout
        layout = QtWidgets.QGridLayout()
        layout_row = 0

        # button to turn the laser off
        self.off_button = QtWidgets.QPushButton('Off')
        self.off_button.clicked.connect(lambda: self.laser.turn_off())
        self.off_button.clicked.connect(self.update_state)
        layout.addWidget(self.off_button, layout_row, 0)

        # button to turn the laser on
        self.on_button = QtWidgets.QPushButton('On')
        self.on_button.clicked.connect(lambda: self.laser.turn_on())
        self.on_button.clicked.connect(self.update_state)
        layout.addWidget(self.on_button, layout_row, 1)
        layout_row += 1

        # power spinbox
        layout.addWidget(QtWidgets.QLabel('Power'), layout_row, 0)
        self.power_spinbox = SpinBox(value=1e-3, suffix='W', siPrefix=True, bounds=(0, 100e-3), dec=True, minStep=0.1e-6)
        def set_power(spinbox):
            self.laser.set_power(spinbox.value()*1000)
            self.laser.set_modulation_power(spinbox.value()*1000)
        self.power_spinbox.sigValueChanged.connect(set_power)
        self.power_spinbox.setValue(value=0)
        layout.addWidget(self.power_spinbox, layout_row, 1)
        layout_row += 1

        # dropdown menu to pick laser mode
        layout.addWidget(QtWidgets.QLabel('Mode'), layout_row, 0)
        self.mode_dropdown = QtWidgets.QComboBox()
        self.mode_dropdown.addItem('Digital Modulation') # index 0
        self.mode_dropdown.addItem('Constant Power') # index 1
        # default to modulation mode
        self.mode_dropdown.setCurrentIndex(0)
        self.mode_dropdown.currentIndexChanged.connect(self.change_mode)
        layout.addWidget(self.mode_dropdown, layout_row, 1)
        layout_row += 1

        # state label
        self.state_label = QtWidgets.QLabel('')
        self.update_state()
        layout.addWidget(self.state_label, layout_row, 1)
        # get state button
        self.get_state_button = QtWidgets.QPushButton('Get State')
        self.get_state_button.clicked.connect(self.update_state)
        layout.addWidget(self.get_state_button, layout_row, 0)
        layout_row += 1

        # take up any additional space in the final column with padding
        layout.setColumnStretch(2, 1)
        # take up any additional space in the final row with padding
        layout.setRowStretch(layout_row, 1)

        self.setLayout(layout)

    def update_state(self):
        """Query the laser for the current state then update the state text box."""
        self.state_label.setText(self.laser.get_state()[4:])

    def change_mode(self, idx):
        """Change the laser output mode."""
        # constant power
        # digital modulation
        if idx == 0:
            self.laser.modulation_mode()
        elif idx == 1:
            self.laser.constant_power()
        else:
            raise ValueError('Invalid mode selection')
        self.update_state()



"""
GUI for Arduino GPIO.

Copyright (c) 2022, Jacob Feder
All rights reserved.
"""
from functools import partial

from pyqtgraph.Qt import QtWidgets

class ArduinoGPIOWidget(QtWidgets.QWidget):
    """Qt widget for controlling the Arduino GPIO."""
    def __init__(self, arduino):
        """
        Args:
            arduino: The Arduino driver.
        """
        super().__init__()

        self.arduino = arduino

        # top level layout
        layout = QtWidgets.QGridLayout()
        layout_row = 0

        # DO state dropdowns
        self.digital_out_dropdowns = {}
        # run pin_changed_external() when the pin state is changed from another source
        self.arduino.pin_state_changed.connect(self.pin_changed_external)
        for i, pin in enumerate(self.arduino.pins):
            layout.addWidget(QtWidgets.QLabel(f'Digital Out {pin}'), layout_row, 0)
            self.digital_out_dropdowns[pin] = QtWidgets.QComboBox()
            self.digital_out_dropdowns[pin].addItem('Low') # index 0
            self.digital_out_dropdowns[pin].addItem('High') # index 1
            # default to low
            self.digital_out_dropdowns[pin].setCurrentIndex(0)
            # update the driver state when the combobox is changed
            self.digital_out_dropdowns[pin].currentIndexChanged.connect(partial(self.combobox_changed, pin))
            layout.addWidget(self.digital_out_dropdowns[pin], layout_row, 1)
            layout_row += 1

        # take up any additional space in the final column with padding
        layout.setColumnStretch(2, 1)
        # take up any additional space in the final row with padding
        layout.setRowStretch(layout_row, 1)

        self.setLayout(layout)

    def pin_changed_external(self, pin, state):
        """The underlying driver changed the pin state"""
        self.digital_out_dropdowns[pin].blockSignals(True)
        if state:
            self.digital_out_dropdowns[pin].setCurrentIndex(1)
        else:
            self.digital_out_dropdowns[pin].setCurrentIndex(0)
        self.digital_out_dropdowns[pin].blockSignals(False)

    def combobox_changed(self, pin, index):
        """User changed the combobox selection"""
        self.arduino.set_pin(pin, index, emit=False)



"""Wrapper driver for NI-DAQ modules. Implements some basic functionality to make the DAQ easier to use.

nidaqmx low-level driver needs to be installed first: https://www.ni.com/en-us/support/downloads/drivers/download.ni-daqmx.html

For installation on Ubunutu:
- Not all hardware is compatible, first check: https://www.ni.com/en-us/support/documentation/compatibility/21/ni-hardware-and-operating-system-compatibility.html
- Download https://www.ni.com/en-us/support/downloads/drivers/download.ni-daqmx.html
- extract the zip file
- install the "drivers" package e.g.: sudo apt install ./ni-ubuntu2204firstlook-drivers-stream.deb
- install the ni-daqmx package: sudo apt install ni-daqmx (may need to run sudo apt update)
- sudo dkms autoinstall
- (BEN: I needed to run python -m pip install nidaqmx)
- reboot
"""

import logging
import nidaqmx

logger = logging.getLogger(__name__)

class DAQ:
    def __init__(self, serial_num=None):
        """
        Args:
            serial_num: device serial number as a hex string e.g. '20BA17A' - leave as None to connect to any available nidaqmx device
        """
        self.system = nidaqmx.system.System.local()
        self.dev = None
        # iterate over the nidaqmx devices
        for d in self.system.devices:
            logger.debug(f'Found nidaqmx device [{d.name}] with serial number [{self.dec_sn_to_hex(d.dev_serial_num)}]')
            if serial_num is None or d.dev_serial_num == self.hex_sn_to_dec(serial_num):
                self.dev = d
        # check whether the device was found
        if self.dev is not None:
            logger.info(f'Bound to nidaqmx device [{self.dev.name}] with serial number [{self.dec_sn_to_hex(self.dev.dev_serial_num)}]')
        else:
            if serial_num is None:
                raise ValueError('No nidaqmx devices found.')
            else:
                raise ValueError(f'No nidaqmx device with the serial number {serial_num} was found.')
        self.task = None

    def __repr__(self):
        if self.dev is not None:
            return f'DAQ(name={self.dev.name}, serial_num={self.dec_sn_to_hex(self.dev.dev_serial_num)})'
        else:
            return 'DAQ(Not connected)'

    def serial_number(self):
        """Return the device serial number as a hex string"""
        return self.dec_sn_to_hex(self.dev.dev_serial_num)

    def dec_sn_to_hex(self, dec):
        """Convert the device serial number from decimal to hex form."""
        return hex(dec)[2:].upper()

    def hex_sn_to_dec(self, hex):
        """Convert the device serial number from hex to decimal form."""
        return int(hex, base=16)

    def reset(self):
        self.dev.reset_device()

    def _remove_dev_prefix(self, l):
        """remove device name e.g. 'Dev1/' prefix from all elements in list l."""
        trimmed_l = []
        for c in l:
            trimmed_l.append(c.split('/', 1)[1])
        return trimmed_l

    def ao_chans(self):
        """Return all of the analog output channel names e.g. ['ao0', 'ao1', ...]"""
        return self._remove_dev_prefix(self.dev.ao_physical_chans.channel_names)

    def ai_chans(self):
        """Return all of the analog input channel names e.g. ['ai0', 'ai1', ...]"""
        return self._remove_dev_prefix(self.dev.ai_physical_chans.channel_names)

    def do_lines(self):
        """Return all of the digital output line names e.g. ['port0/line0', 'port0/line1', ...]"""
        return self._remove_dev_prefix(self.dev.do_lines.channel_names)

    def set_ao(self, ch, val):
        """Set the analog output voltage for a given channel.

        Args:
            ch: analog output channel string e.g. 'ao0', 'ao1', ...
            val: analog output value (V)
        """
        with nidaqmx.Task() as task:
            task.ao_channels.add_ao_voltage_chan(f'{self.dev.name}/{ch}')
            task.write(val, auto_start=True)

    def get_ai_se(self, ch):
        """Get the single-ended analog input voltage on a given channel.

        Args:
            ch: analog input channel string e.g. 'ai0', 'ai1', ...
        """
        ret = None
        with nidaqmx.Task() as task:
            task.ai_channels.add_ai_voltage_chan(f'{self.dev.name}/{ch}', terminal_config=nidaqmx.constants.TerminalConfiguration.RSE)
            ret = task.read()
        return ret

    def set_do(self, port, line, state):
        """Set a digital output voltage.

        Args:
            port: device port string e.g. 'port0', 'port1', ...
            line: port line string e.g. 'line0', 'line1', ...
            state: output state boolean
        """
        with nidaqmx.Task() as task:
            task.do_channels.add_do_chan(f'{self.dev.name}/{port}/{line}')
            task.write(state)

if __name__ == '__main__':
    import time
    # enable logging to console
    logging.basicConfig(level=logging.DEBUG, format='%(asctime)s.%(msecs)03d [%(levelname)8s] %(message)s', datefmt='%m-%d-%Y %H:%M:%S')
    daq = DAQ(serial_num='20BA17A')
    import pdb; pdb.set_trace()
    for i in range(3):
        daq.set_ao('ao1', 1.2)
        daq.set_do('port0', 'line1', True)
        time.sleep(1)
        print(f'ai voltage (on): {daq.get_ai_se("ai0")}')
        daq.set_ao('ao1', 0.0)
        daq.set_do('port0', 'line1', False)
        time.sleep(1)
        print(f'ai voltage (off): {daq.get_ai_se("ai0")}')



"""Rohde and Schwarz HMP4040 driver.

Programming manual:
https://scdn.rohde-schwarz.com/ur/pws/dl_downloads/dl_common_library/dl_manuals/gb_1/h/hmp_serie/HMPSeries_UserManual_en_02.pdf

Copyright (c) 2022, Jacob Feder
All rights reserved.
"""

import time
import logging

from pyvisa import ResourceManager

logger = logging.getLogger(__name__)

class HMP4040:
    def __init__(self, address):
        """
        Args:
            address: PyVISA resource path.
        """
        self.address = address
        self.rm = ResourceManager('@py')

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *args):
        self.close()

    def __str__(self):
        return f'{self.address} {self.idn}'

    def open(self):
        try:
            self.device = self.rm.open_resource(self.address)
        except Exception as err:
            raise ConnectionError(f'Failed connecting to HMP4040 @ [{self.address}]') from err
        self.device.read_termination = '\n'
        self.device.write_termination = '\n'
        # 1 second timeout
        self.device.timeout = 1000
        self.idn = self.device.query('*IDN?')
        if int(self.device.query('*TST?')):
            raise RuntimeError(f'HMP4040 [{self}] self-test failed.')
        logger.info(f'Connected to HMP4040 [{self}].')
        return self

    def close(self):
        self.device.close()

    def _write(self, cmd):
        """Send a VISA command to the device.
        Args:
            cmd: command string to send
        """
        self.device.write(cmd)
        # wait until the command was processed
        self.device.write('*WAI')
        # check for errors
        # query('SYST:ERR?') returns a string like '-100, "Command error"'
        err = self.device.query('SYST:ERR?').replace('"', '').split(', ')
        err_num = int(err[0])
        if err_num:
            # empty error queue
            other_errs = err_num
            while other_errs:
                other_errs = int(self.device.query('SYST:ERR?').split(', ')[0])
            raise RuntimeError(f'HMP4040 communication error: {err}')

    def _select_output(self, ch):
        """Select the channel that will receive subsequent commands.

        Args:
            ch: output channel (e.g. 1, 2, 3)
        """
        self._write(f'INST:NSEL {ch}')

    def set_output(self, ch, state):
        """Turn the channel output on or off.

        Args:
            ch: output channel (e.g. 1, 2, 3)
            state: True to enable the channel, False to disable
        """
        self._select_output(ch)
        if state:
            self._write(f'OUTP:STAT ON')
        else:
            self._write(f'OUTP:STAT OFF')

    def set_voltage(self, ch, val, confirm=True, timeout=1.0, delta=0.05):
        """Set the channel voltage.

        Args:
            ch: output channel (e.g. 1, 2, 3)
            val: channel output (volts)
            confirm: measure the voltage continuously until it is within 
                    delta (volts) of the set voltage.
            timeout: max allowed time (s) to reach the set voltage
            delta: acceptable delta from set voltage (volts)
        """
        self._select_output(ch)
        self._write(f'VOLT {val}')
        if confirm:
            timeout = time.time() + timeout
            actual = self.measure_voltage(ch=ch)
            while abs(val - actual) > delta:
                time.sleep(0.1)
                if time.time() > timeout:
                    raise TimeoutError(f'Measured channel {ch} voltage [{actual}] did not reach set voltage [{val}].')
                actual = self.measure_voltage(ch=ch)

    def set_current(self, ch, val, confirm=True, timeout=1.0, delta=0.05):
        """Set the channel current.

        Args:
            ch: output channel (e.g. 1, 2, 3)
            val: channel output (amps)
            confirm: measure the current continuously until it is within 
                    delta (amps) of the set current.
            timeout: max allowed time (s) to reach the set current
            delta: acceptable delta from set current (amps)
        """
        self._select_output(ch)
        self._write(f'CURR {val}')
        if confirm:
            timeout = time.time() + timeout
            actual = self.measure_current(ch=ch)
            while abs(val - actual) > delta:
                time.sleep(0.1)
                if time.time() > timeout:
                    raise TimeoutError(f'Measured channel {ch} current [{actual}] did not reach set current [{val}].')
                actual = self.measure_current(ch=ch)

    def set_ovp(self, ch, val):
        """Set the channel over-voltage protection trip threshold.

        Args:
            ch: output channel (e.g. 1, 2, 3)
            val: channel ovp limit
        """
        self._select_output(ch)
        self._write(f':VOLT:PROT:MODE MEAS')
        self._write(f'VOLT:PROT {val}')

    def clear_ovp(self, ch):
        """Clear and disable the channel over-voltage protection.

        Args:
            ch: output channel (e.g. 1, 2, 3)
        """
        self._select_output(ch)
        self._write(f':VOLT:PROT:CLE')

    def measure_voltage(self, ch):
        """Return the actual channel voltage.

        Args:
            ch: output channel (e.g. 1, 2, 3)

        Returns:
            Channel voltage as a float 
        """
        self._select_output(ch)
        volt = float(self.device.query(f'MEAS:VOLT?'))
        return volt

    def measure_current(self, ch):
        """Return the actual channel current.

        Args:
            ch: output channel (e.g. 1, 2, 3)

        Returns:
            Channel current as a float 
        """
        self._select_output(ch)
        curr = float(self.device.query(f'MEAS:CURR?'))
        return curr



"""Beaglebone Black GPIO driver (beaglebone must be running flask server driver).

Author: Jacob Feder
Date: 8/5/2022
"""
import requests
import logging
import time

logger = logging.getLogger(__name__)

try:
    from pyqtgraph.Qt import QtCore
    QObject = QtCore.QObject
except ModuleNotFoundError:
    logger.info('Not declaring GUI because the required packages are not available')
    QObject = object
    gui = False
else:
    gui = True

class BBGPIO(QObject):
    if gui:
        # for when the pin state is changed
        # args: pin id, state
        pin_state_changed = QtCore.Signal(str, bool)

    def __init__(self, hostname='beaglebone.local', port=5000, timeout=5):
        super().__init__()
        self.hostname = hostname
        self.port = port
        self.timeout = timeout

        # available pins
        # see https://vadl.github.io/beagleboneblack/2016/07/29/setting-up-bbb-gpio
        self.pins = []
        # pins on P8 header
        for i in range(7, 20):
            self.pins += [f'P8_{i}']
        self.pins += [f'P8_26']
        # pins on P9 header
        for i in range(11, 19):
            self.pins += [f'P9_{i}']
        for i in range(21, 25):
            self.pins += [f'P9_{i}']
        self.pins += ['P9_26', 'P9_27']

    def _request(self, msg):
        """Send a http request to the flask server.
        Args:
            msg: the https address e.g. '/gpio/output/high/<string:terminal>'

        Returns:
            http resonse text.

        Raises:
            ConnectionError: if the request failed
        """
        logger.debug(f'bbgpio request: [{msg}]')
        try:
            ret = requests.get(f'http://{self.hostname}:{self.port}{msg}', timeout=self.timeout)
            if ret.status_code != 200:
                raise ConnectionError
        except (requests.exceptions.ConnectionError, ConnectionError) as e :
            raise ConnectionError('flask request failed') from e
        logger.debug(f'bbgpio response: [{ret.text}]')
        return ret.text

    def connected(self):
        """Check if the server is accessible and running.

        Returns:
            True if so, False otherwise.
        """
        try:
            self._request(f'/')
        except ConnectionError:
            return False
        else:
            logger.info(f'Connected to Beaglebone@{self.hostname}:{self.port}')
            return True

    def restart(self):
        """Restart the flask service on the beaglebone.

        Raises:
            ConnectionError: if the beaglebone isn't responding.
        """

        # trigger the restart
        try:
            self._request('/restart')
        except ConnectionError:
            pass

        # wait until the service is back up
        timeout = time.time() + 5.0
        while True:
            if time.time() > timeout:
                raise ConnectionError('timed out waiting for beaglebone to restart')
            try:
                self._request('')
            except ConnectionError:
                pass
            else:
                break
            time.sleep(0.1)

    def set_pin(self, pin, state, emit=True):
        """Set the output state of a GPIO pin on the Beaglebone.

        Args:
            pin: pin name string like 'P8_35'
            state: True for high, False for low
            emit: True to emit a PyQt signal when the function finishes

        Raises:
            ValueError: invalid pin state
            RunTimeError: flask failure
        """
        if pin not in self.pins:
            raise ValueError(f'Pin is invalid or already allocated for another function. Available pins:\n{self.pins}')
        if state:
            self._request(f'/gpio/output/high/{pin}')
            logger.debug(f'Set Beaglebone@{self.hostname}:{self.port} pin [{pin}] HIGH')
        else:
            self._request(f'/gpio/output/low/{pin}')
            logger.debug(f'Set Beaglebone@{self.hostname}:{self.port} pin [{pin}] LOW')
        if gui and emit:
            self.pin_state_changed.emit(pin, state)

if __name__ == '__main__':
    bbgpio = BBGPIO()
    bbgpio.restart()
    bbgpio.set_pin('P8_14', False)



"""
GUI for a generic camera accessible from openCV.

Copyright (c) 2022, Jacob Feder
All rights reserved.
"""
from functools import partial
import logging
from threading import Lock

import cv2
import numpy as np
from pyqtgraph.Qt import QtGui
from pyqtgraph.Qt import QtCore
from pyqtgraph.Qt import QtWidgets
from pyqtgraph import SpinBox


class VideoStreamer(QtCore.QObject):
    """Worker object to stream data from the camera in a thread."""
    new_frame = QtCore.Signal(np.ndarray)

    def __init__(self, cam):
        super().__init__()
        self.cam = cam
        self.acquiring = False
        self.lock = Lock()
        self.sem = QtCore.QSemaphore(n=1)

    def acquire(self):
        """Continuously acquire images."""
        with self.lock:
            if self.acquiring:
                logger.warning('Cannot start acquisition because camera is already acquiring.')
                return
            self.acquiring = True
        while self.acquiring:
            # don't get the next frame until the previous frame has been displayed
            # in order to prevent overwhelming the GUI
            self.sem.acquire()
            ret, frame = self.cam.read()
            if not ret:
                raise RuntimeError('Failed reading camera frame')
            self.new_frame.emit(frame)


class CV2CameraWidget(QtWidgets.QWidget):
    """Qt widget for displaying the camera feed."""
    def __init__(self, camera_port=0):
        """
        Args:
            camera_port: the openCV port number of the camera (0 for first connected camera, 1 for second, ...)
        """
        super().__init__()

        # create camera driver
        self.cam = cv2.VideoCapture(camera_port)
        if not self.cam.isOpened():
            raise RuntimeError('Failed connecting to camera.')

        # top level layout
        layout = QtWidgets.QVBoxLayout()

        self.disply_width = 640
        self.display_height = 480
        # create the label that holds the image
        self.image_label = QtWidgets.QLabel()
        self.image_label.resize(self.disply_width, self.display_height)
        layout.addWidget(self.image_label)

        # worker object to read the camera data
        self.stream_worker = VideoStreamer(self.cam)
        # proper Qt thread handling
        # https://mayaposch.wordpress.com/2011/11/01/how-to-really-truly-use-qthreads-the-full-explanation/
        self.thread = QtCore.QThread()
        self.stream_worker.moveToThread(self.thread)
        # stop the streamer when this object is destroyed
        self.destroyed.connect(partial(self.stop_streamer))
        self.destroyed.connect(self.thread.quit)
        self.thread.finished.connect(partial(self.exit))
        # run update_image to update the QLabel whenever a new image is available from the camera
        self.stream_worker.new_frame.connect(self.update_image)
        # start the thread
        self.thread.start()

        # acuisition controls
        controls_layout = QtWidgets.QHBoxLayout()

        # button to turn the acquisition on
        self.start_button = QtWidgets.QPushButton('Start')
        self.start_button.clicked.connect(self.stream_worker.acquire)
        controls_layout.addWidget(self.start_button)

        # button to turn the acquisition off
        self.stop_button = QtWidgets.QPushButton('Stop')
        self.stop_button.clicked.connect(self.stop_streamer)
        controls_layout.addWidget(self.stop_button)

        # gain spinbox
        controls_layout.addWidget(QtWidgets.QLabel('Gain'))
        self.gain_spinbox = SpinBox(value=1, bounds=(0, 2e16), dec=True, minStep=1e-3)
        self.gain_spinbox.sigValueChanged.connect(lambda spinbox: self.set_gain(spinbox.value()))
        self.gain_spinbox.setValue(1000)
        controls_layout.addWidget(self.gain_spinbox)

        layout.addLayout(controls_layout)
        self.setLayout(layout)

    def stop_streamer(self):
        self.stream_worker.acquiring = False

    def exit(self):
        self.cam.release()

    def set_gain(self, gain):
        self.gain = gain

    def update_image(self, frame):
        """Updates the image_label with a new opencv image"""
        qt_img = self.convert_cv_qt(frame)
        self.image_label.setPixmap(qt_img)
        self.stream_worker.sem.release()

    # https://gist.github.com/docPhil99/ca4da12c9d6f29b9cea137b617c7b8b1
    def convert_cv_qt(self, img):
        """Convert from an opencv image to QPixmap"""
        rgb_image = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        h, w, ch = rgb_image.shape
        bytes_per_line = ch * w
        convert_to_Qt_format = QtGui.QImage(rgb_image.data, w, h, bytes_per_line, QtGui.QImage.Format.Format_RGB888)
        p = convert_to_Qt_format.scaled(self.disply_width, self.display_height, QtCore.Qt.AspectRatioMode.KeepAspectRatio)
        return QtGui.QPixmap.fromImage(p)



"""Rigol DP832 driver.

Based on https://github.com/kearneylackas/DP832-Python

Programming manual:
https://www.batronix.com/pdf/Rigol/ProgrammingGuide/DP800_ProgrammingGuide_EN.pdf

Copyright (c) 2022, Jacob Feder
All rights reserved.
"""

import time
import logging

from pyvisa import ResourceManager

logger = logging.getLogger(__name__)

class DP832:
    def __init__(self, address):
        """
        Args:
            address: PyVISA resource path.
        """
        # time to wait (s) after each command
        self.delay = 0.01
        self.rm = ResourceManager('@py')
        self.address = address

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *args):
        self.close()

    def __str__(self):
        return f'{self.address} {self.idn}'

    def open(self):
        try:
            self.device = self.rm.open_resource(self.address)
        except Exception as err:
            raise ConnectionError(f'Failed connecting to DP832 @ [{self.address}]') from err
        # 1 second timeout
        self.device.timeout = 1000
        self.idn = self.device.query('*IDN?')
        logger.info(f'Connected to DP832 [{self}].')

        return self

    def close(self):
        self.device.close()

    def _write(self, cmd):
        """Send a VISA command to the device.
        Args:
            cmd: command string to send
        """
        self.device.write(cmd)
        # sleep to allow time for the command to take effect
        time.sleep(self.delay)

    def _select_output(self, ch):
        """Select the channel that will receive subsequent commands.

        Args:
            ch: output channel (e.g. 1, 2, 3)
        """
        self._write(f':INST:NSEL {ch}')

    def toggle_output(self, ch, state):
        """Turn the channel output on or off.

        Args:
            ch: output channel (e.g. 1, 2, 3)
            state: True to enable the channel, False to disable
        """
        if state:
            self._write(f':OUTP CH{ch},ON')
        else:
            self._write(f':OUTP CH{ch},OFF')

    def set_voltage(self, ch, val, confirm=True, timeout=1.0, delta=0.03):
        """Set the channel voltage.

        Args:
            ch: output channel (e.g. 1, 2, 3)
            val: channel output (volts)
            confirm: measure the voltage continuously until it is within 
                    delta (volts) of the set voltage.
            timeout: max allowed time (s) to reach the set voltage
            delta: acceptable delta from set voltage (volts)
        """
        self._select_output(ch)
        self._write(f':VOLT {val}')
        if confirm:
            timeout = time.time() + timeout
            actual = self.measure_voltage(ch=ch)
            while abs(val - actual) > delta:
                time.sleep(0.1)
                if time.time() > timeout:
                    raise TimeoutError(f'Measured channel {ch} voltage [{actual}] did not reach set voltage [{val}].')
                actual = self.measure_voltage(ch=ch)

    def set_current(self, ch, val, confirm=True, timeout=1.0, delta=0.02):
        """Set the channel current.

        Args:
            ch: output channel (e.g. 1, 2, 3)
            val: channel output (amps)
            confirm: measure the current continuously until it is within 
                    delta (amps) of the set current.
            timeout: max allowed time (s) to reach the set current
            delta: acceptable delta from set current (amps)
        """
        self._select_output(ch)
        self._write(f':CURR {val}')
        if confirm:
            timeout = time.time() + timeout
            actual = self.measure_current(ch=ch)
            while abs(val - actual) > delta:
                time.sleep(0.1)
                if time.time() > timeout:
                    raise TimeoutError(f'Measured channel {ch} current [{actual}] did not reach set current [{val}].')
                actual = self.measure_current(ch=ch)

    def set_ovp(self, ch, val):
        """Set the channel over-voltage protection.

        Args:
            ch: output channel (e.g. 1, 2, 3)
            val: channel ovp limit
        """
        self._select_output(ch)
        self._write(f':VOLT:PROT {val}')

    def toggle_ovp(self, ch, state):
        """Enable or disable the channel over-voltage protection.

        Args:
            ch: output channel (e.g. 1, 2, 3)
            state: True to enable the OVP, False to disable
        """
        self._select_output(ch)
        if state:
            self._write(f':VOLT:PROT:STAT ON')
        else:
            self._write(f':VOLT:PROT:STAT OFF')

    def set_ocp(self, ch, val):
        """Set the channel over-current protection.

        Args:
            ch: output channel (e.g. 1, 2, 3)
            val: channel ocp limit
        """
        self._select_output(ch)
        self._write(f':CURR:PROT {val}')

    def toggle_ocp(self, ch, state):
        """Enable or disable the channel over-current protection.

        Args:
            ch: output channel (e.g. 1, 2, 3)
            state: True to enable the OCP, False to disable
        """
        self._select_output(ch)
        if state:
            self._write(f':CURR:PROT:STAT ON')
        else:
            self._write(f':CURR:PROT:STAT OFF')

    def measure_voltage(self, ch):
        """Return the actual channel voltage.

        Args:
            ch: output channel (e.g. 1, 2, 3)

        Returns:
            Channel voltage as a float 
        """
        volt = float(self.device.query(f':MEAS:VOLT? CH{ch}'))
        time.sleep(self.delay)
        return volt

    def measure_current(self, ch):
        """Return the actual channel current.

        Args:
            ch: output channel (e.g. 1, 2, 3)

        Returns:
            Channel current as a float 
        """
        curr = float(self.device.query(f':MEAS:CURR? CH{ch}'))
        time.sleep(self.delay)
        return curr

    def measure_power(self, ch):
        """Return the actual channel power.

        Args:
            ch: output channel (e.g. 1, 2, 3)

        Returns:
            Channel power as a float 
        """
        power = float(self.device.query(f':MEAS:POWE? CH{ch}'))
        time.sleep(self.delay)
        return power



"""LuminOX LOX-02 oxygen sensor driver.
https://gaslab.com/products/oxygen-sensor-luminox-lox-o2

Author: Jacob Feder, Kimberly Carrillo Rivera
Date: 8/12/2021
"""

import serial
import logging
import time

logger = logging.getLogger(__name__)

class LOX_02:
    def __init__(self, serial_port: str):
        """Args:
            serial_port: serial COM port (see pyserial docs)
        """
        self.serial_port = serial_port

    def open(self):
        self.serial_port = serial.Serial(self.serial_port)
        self.mode_poll()
        self.serial_port.reset_input_buffer()

    def close(self):
        self.serial_port.close()

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *args):
        self.close()

    def _format_msg(self, command, arg=None):
        msg = command
        if arg:
            msg = msg + ' ' + arg
        msg += '\r\n'
        logger.debug(f'sent message [{msg}]')
        self.serial_port.write(msg.encode('ascii'))

    def mode_poll(self):
        self._format_msg('M', arg='1')
        response = self.serial_port.readline() # 'M 01\r\n'

    def read_all(self):
        self.mode_poll() # 'M 1\r\n'
        self._format_msg('A') # 'A\r\n'
        response = self.serial_port.readline() # 'O xxxx.x T yxx.x P xxxx % xxx.xx e xxxx\r\n'
        logger.debug(f'response [{response}]')
        ppo2 = float(response[2:8])
        temp = float(response[11:16])
        pressure = float(response[19:23])
        percento2 = float(response[26:32])
        status = int(response[35:39])
        return (ppo2, temp, pressure, percento2, status)

    def ppo2(self):
        """Return the partial pressure of oxygen (mbar)"""
        ppo2, temp, pressure, percento2, status = self.read_all()
        return ppo2
    
    def temp(self):
        """Return the temperature (deg C)"""
        ppo2, temp, pressure, percento2, status = self.read_all()
        return temp

    def o2(self):
        """Return the O2 concentration as a percentage"""
        ppo2, temp, pressure, percento2, status = self.read_all()
        return percento2

    def pressure(self):
        """Return the total pressure in mbar"""
        ppo2, temp, pressure, percento2, status = self.read_all()
        return pressure



"""
Simple wrapper interface for Zaber stages.

Copyright (c) 2022, Jacob Feder
All rights reserved.
"""
import logging
from pathlib import Path
from collections import OrderedDict

from zaber_motion import Library, Units
from zaber_motion.ascii import Connection
Library.enable_device_db_store()

logger = logging.getLogger(__name__)

class ZaberStages:
    """Collection of Zaber stages. Automatically connects to the stages and allows intuitive access to the underlying zaber axis objects."""
    def __init__(self, axis_mapping, serial_port=None):
        """
        Args:
            serial_port: serial COM port (see pyserial docs). Leave as None to detect automatically (linux only).
            axis_mapping: dictionary mapping an axis peripheral id (int) to the axis name (str)
        """
        self.axis_mapping = axis_mapping
        self.serial_port = serial_port

    def __getitem__(self, key):
        """Return the zaber axis object associated with the given key."""
        if key in self.axes:
            return self.axes[key]
        else:
            raise KeyError(f'Zaber axis key [{key}] not found')

    def __iter__(self):
        return self.axes.__iter__()

    def __next__(self):
        return self.axes.__next__()

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *args):
        self.close()

    def open(self):
        if self.serial_port is not None:
            self.conn = Connection.open_serial_port(self.serial_port)
            logger.debug(f'Found {len(device_list)} devices')
        else:
            # detect any connected Zaber serial devices
            zaber_serial_ports = Path('/dev/serial/by-id/').glob('usb-Zaber_Technologies_Inc.*')
            self.conn = None
            for s in zaber_serial_ports:
                # connect to the first one and then exit the loop
                self.conn = Connection.open_serial_port(str(s))
                break
            if self.conn is None:
                raise IOError('No Zaber devices found')

        self.devices = self.conn.detect_devices()
        logger.debug(f'Found [{len(self.devices)}] zaber devices')

        self.axes = OrderedDict()
        for device in self.devices:
            for i in range(1, device.axis_count + 1):
                axis = device.get_axis(i)
                logger.debug(f'Found Zaber axis [{axis}] with peripheral id [{axis.peripheral_id}]')
                if axis.peripheral_id in self.axis_mapping:
                    axis_key = self.axis_mapping[axis.peripheral_id]
                    logger.debug(f'Associated zaber axis [{axis_key}] with peripheral id [{axis.peripheral_id}]')
                    self.axes[axis_key] = axis

    def close(self):
        self.conn.close()



"""DLnsec laser driver.

https://labs-electronics.com/

Author: Jacob Feder, Will Burke
Date: 2/22/2022
"""
import time
import logging

import serial

logger = logging.getLogger(__name__)

class DLnsec():
    def __init__(self, serial_port: str, timeout=1.0):
        """
        Args:
            serial_port: serial COM port (see pyserial docs)
        """
        self.serial_port = serial_port
        self.timeout = timeout

    def open(self):
        """Connect to the laser."""
        self.conn = serial.Serial(self.serial_port, baudrate=9600, timeout=self.timeout)
        self.conn.reset_input_buffer()
        self.conn.reset_output_buffer()
        self.reboot()
        # query the laser serial number
        self.idn = self._query('*IDN')
        if not self.idn:
            raise RuntimeError('Failed getting laser ID number')
        logger.info(f'Connected to DLnsec [{self.idn}]')

    def close(self):
        """Disconnect from the laser."""
        self.conn.close()

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *args):
        self.off()
        self.close()

    def _send_raw(self, msg):
        """Send a command to the laser."""
        # add a newline and encode the message into a binary string
        msg_bin = (msg + '\n').encode('ascii')
        # send the message on the serial port
        self.conn.write(msg_bin)
        logger.debug(f'Sent message to DLnsec: [{msg}]')
        # wait for the laser to process the message
        time.sleep(0.1)

    def _send(self, msg):
        """Send a command to the laser and check for errors."""
        # number of times to try sending the command
        attempts = 3
        while attempts:
            # send the command
            self._send_raw(msg)
            # ask if an error occurred
            response = self._query('ERR?')
            if response == '0; No error':
                break
            else:
                attempts -= 1
                err_str = f'Failed sending DLnsec command [{msg}]: response [{response}]'
                if attempts:
                    logger.warning(err_str)
                else:
                    raise RuntimeError(err_str)
                # wait a moment before trying again
                time.sleep(0.25)

    def _query(self, msg):
        """Send a query command to the laser and return the response."""
        self._send_raw(msg)
        # receive the response
        response = self.conn.read_until(b'\n\r')
        response = response.decode('ascii').strip('\n\r')
        logger.debug(f'Response from DLnsec: [{response}]')

        return response

    def power(self, p):
        """Set the output power (integer percentage)"""
        if not isinstance(p, int) or p > 100 or p < 0:
            raise ValueError(f'The power [{p}] must be an integer between 0-100')
        self._send(f'PWR{p}')

    def cw_mode(self):
        """Put the laser into CW output mode."""
        self._send('LASE')

    def on(self):
        """Turn the output stage on."""
        self._send('*ON')

    def off(self):
        """Turn the output stage off."""
        self._send('*OFF')

    def reboot(self):
        """Reboot the laser, applying default settings on startup."""
        self._send_raw('*RBT')
        # wait for the laser to fully restart
        time.sleep(2)



"""
Driver for the Ocean Optics spectrometer HR2000+ES
Author: Benjamin Soloway
Date: 09/14/2022

Use spec.features to see cseabreeze features
Information on this is in python-seabreeze.readthedocs.io/en/latest/backend_api.html
Access trigger mode using
spec.features['spectrometer'][0].set_trigger_mode(5)

Based on page 18 of this reference
https://www.oceaninsight.com/globalassets/catalog-blocks-and-images/manuals--instruction-old-logo/electronic-accessories/external-triggering-options_firmware3.0andabove.pdf
It looks like we want to use data value 4 which will turn on external hardware edge trigger mode.
* Note from page 14 that it looks like it takes 900us to put data into FIFO and 1ms to read data from FIFO.

"In the External Hardware Edge Trigger mode, a rising edge detected by the FPGA from the External
Trigger input starts the Integration Cycle specified through the software interface. After the Integration
Cycle completes, the spectrum is retrieved and written to the FIFO in the FPGA followed by a CCD Reset
Cycle. Only one acquisition will be performed for each External Trigger pulse, no matter what the
pulse’s duration is. The Reset Cycle insures that the CCD performance uniform on a scan-to-scan basis.
The time duration for this reset cycle is relative to the Integration Cycle time and will change if the
integration period is changed. So the timing sequence is Trigger, Trigger Delay, Integration Cycle,
Read/Write Cycle, Reset Cycle, and Idle Cycle(s). The Idle Cycle will until the next trigger occurs."

From page 19 of https://www.oceaninsight.com/globalassets/catalog-blocks-and-images/manuals--instruction-old-logo/spectrometer/hr2000-.pdf
Looks like on the 30-pin cable we need to trigger on 10 and ground on 5.
On the 15-pin cable we need to trigger on 4,5, or 8 and connect to ground on 10.

Maybe can also set the trigger mode this way?
https://python-seabreeze.readthedocs.io/en/latest/api.html?highlight=set_trigger_mode#seabreeze.spectrometers.Spectrometer.trigger_mode
"""
# import multiprocessing

import seabreeze.spectrometers as sb
import time
import logging
try:
    from rpyc.utils.classic import obtain
except ModuleNotFoundError:
    def obtain(i):
        return i

logger = logging.getLogger(__name__)

class HR2000ES():
    def __init__(self):
        self.acq_start_time = 0
        self.elapsed_time_sec = 0

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *args):
        self.spec.close()

    def open(self):
        self.spec = sb.Spectrometer.from_serial_number()

    def set_int_time_sec(self, integration_time):
        """
        set the acquisition integration time in seconds
        """
        self.spec.integration_time_micros(integration_time*10**6)

    def get_time_acquiring_sec(self):
        """
        get the time that the spectrometer has spent acquiring
        """
        curr_time = time.time()
        return self.acq_start_time - curr_time

    def get_elapsed_time_sec(self):
        """
        Return time since the last acquisition
        """
        return self.elapsed_time_sec 

    #TODO: make a function to return the time left before acquisition is done. 
    #   This should use a method that gets the integration time from the hardware

    def acquire(self):
        """
        Acquire wavelength data. Record in self.elapsed_time the time it took to acquire the data
        Return the wavelengths, and intensities
        """
        self.acq_start_time = time.time()

        waves = self.spec.wavelengths()
        ints = self.spec.intensities()

        acq_end_time = time.time()
        self.elapsed_time_sec = acq_end_time - self.acq_start_time

        return (waves, ints)

    def acquire_in_process(self, wls, ints):
        """
        Acquire wavelength data. Record in self.elapsed_time the time it took to acquire the data
        I suggest passing wls, ints = [], [] (as
        """
        self.acq_start_time = time.time()

        wls.append(obtain(self.spec.wavelengths()))
        ints.append(obtain(self.spec.intensities()))

        acq_end_time = time.time()
        self.elapsed_time_sec = acq_end_time - self.acq_start_time

    def set_trigger_mode(self, mode):
        # see modes on pg 18 of
        # https://www.oceaninsight.com/globalassets/catalog-blocks-and-images/manuals--instruction-old-logo/electronic-accessories/external-triggering-options_firmware3.0andabove.pdf
        if mode == 0:
            logger.info('Set spectrometer to Normal ("Free Running") Mode')
            self.spec.trigger_mode(mode)
        elif mode == 3:
            logger.info('Set spectrometer to External Hardware Edge Trigger Mode')
            logger.info("An external trigger rising edge will start the acquisition. Only one acquisition will be performed for each External Trigger pulse, no matter what the pulses' duration is.")
            self.spec.trigger_mode(mode)
        else:
            self.spec.trigger_mode(mode)
            logger.info(f"Warning not sure what mode {mode} is.")
            #raise ValueError(f"Invalid mode selection OR if 1,2 I don't know what the mode is: {mode}")

    # def _set_trigger_mode(self, mode):
    #   self.spec.trigger_mode(mode)
    #
    # def _set_trigger(self, mode):
    #   p = multiprocessing.Process(target=self._set_trigger_mode, args=(mode,))
    #   p.start()
    #   p.join(10)
    #   if p.is_alive():
    #       print("running... let's kill it...")
    #       p.terminate()
    #       p.join()

        # elif mode == 1:
        #   logger.info('Set spectrometer to Software Trigger Mode')
        #   logger.info('Continuously takes data when the external trigger is set to high. When low it stops taking data and getSpectrum() will remain blocked (i.e., not return) until the signal goes high again.')
        #   self.spec.trigger_mode(mode)
        # elif mode == 2:
        #   logger.info('Set spectrometer to External Hardware Level Trigger Mode')
        #   logger.info('An external trigger rising edge will start the acquisition. The next acquisition will start if the trigger is still high, FIFO is empty, and spectrum request is active.')
        #   self.spec.trigger_mode(mode)
        # elif mode == 3:
        #   logger.info('Set spectrometer to External Synchronization Trigger Mode')
        #   logger.info('Two triggers needed. The first rising edge starts the integration period and the second rising edge stops the integration while starting the next integration.')
        #   logger.info('Thus the integration time is the period between the two external trigger pulses.')
        #   self.spec.trigger_mode(mode)
        # elif mode == 4:
        #   logger.info('Set spectrometer to External Hardware Edge Trigger Mode')
        #   logger.info("An external trigger rising edge will start the acquisition. Only one acquisition will be performed for each External Trigger pulse, no matter what the pulses' duration is.")
        #   raise ValueError(f'Invalid mode selection: {mode}. The API was wrong when I wrote this code...')
        # else:
        #   raise ValueError(f'Invalid mode selection: {mode}')



"""
Qt GUI for the agiltron fiber switch driver.

Copyright (c) 2022, Jacob Feder, Ben Soloway
All rights reserved.
"""
import logging

from pyqtgraph.Qt import QtWidgets
from pyqtgraph import SpinBox

logger = logging.getLogger(__name__)

class FFSWWidget(QtWidgets.QWidget):
    """Qt widget for controlling cld1010 lasers."""

    def __init__(self, ffsw, port_mapping=None):
        """
        Args:
            ffsw: The fiber-fiber switch (ffsw) driver.
            port_mapping: Dict where the keys are device port numbers and values are string name descriptors
        """
        super().__init__()

        self.ffsw = ffsw

        # top level layout
        layout = QtWidgets.QGridLayout()
        layout_row = 0

        # port dropdown
        layout.addWidget(QtWidgets.QLabel('Port'), layout_row, 0)
        self.port_dropdown = QtWidgets.QComboBox()
        if port_mapping is not None:
            # reverse keys and values
            inverse_port_mapping = {v: k for k, v in port_mapping.items()}
            self.port_dropdown.addItem(inverse_port_mapping[1]) # index 0
            self.port_dropdown.addItem(inverse_port_mapping[2]) # index 1
        else:
            self.port_dropdown.addItem('1') # index 0
            self.port_dropdown.addItem('2') # index 1
        # default to port 1
        self.port_dropdown.setCurrentIndex(0)
        self.port_dropdown.currentIndexChanged.connect(self.change_port)
        layout.addWidget(self.port_dropdown, layout_row, 1)
        layout_row += 1

        # get state button
        self.get_state_button = QtWidgets.QPushButton('Get State')
        self.get_state_button.clicked.connect(self.update_state)
        layout.addWidget(self.get_state_button, layout_row, 0)
        layout_row += 1

        # take up any additional space in the final column with padding
        layout.setColumnStretch(2, 1)
        # take up any additional space in the final row with padding
        layout.setRowStretch(layout_row, 1)

        self.setLayout(layout)

    def change_port(self, idx):
        """Change the agiltron port."""
        if idx == 0:
            self.ffsw.set_port(1)
        elif idx == 1:
            self.ffsw.set_port(2)
        else:
            raise ValueError(f'Invalid mode selection [{idx}]')

    def update_state(self):
        """Query the agiltron for the current state then update the GUI."""
        port = self.ffsw.get_port()
        self.port_dropdown.blockSignals(True)
        if port == 1:
            self.port_dropdown.setCurrentIndex(0)
        elif port == 2:
            self.port_dropdown.setCurrentIndex(1)
        else:
            raise ValueError(f'Got invalid port [{port}]')
        self.port_dropdown.blockSignals(False)



"""Agiltron fiber-fiber switch serial-based driver. Compatible with any switch 
that uses their USB/serial SW-DR-5 control board.

Author: Benjamin Soloway, Jacob Feder
Date: 10/08/2022
"""
import serial
import logging

logger = logging.getLogger(__name__)

class FFSW():
    def __init__(self, serial_port: str):
        self.serial_port = serial_port

    def _read(self):
        response = self.conn.read(4)
        logger.debug(f'response [{response}]')
        return response

    def _write(self, msg):
        self.conn.write(msg)
        logger.debug(f'sent command [{msg}]')

    def get_sn(self):
        self._write(b"\x01\x03\x00\x00")
        a = self._read()[8::] #higher bits
        self._write(b"\x01\x04\x00\x00")
        b = self._read()[8::] #lower bits
        return a+b

    def get_module_address(self):
        #TODO: don't know the meaning of this
        self._write(b"\x01\x01\x00\x00")
        return self._read()

    def set_port(self, port):
        if port == 1:
            self._write(b"\x01\x12\x00\x00")
        elif port == 2:
            self._write(b"\x01\x12\x00\x01")
        logger.info(f'set port to [{port}]')
        return self._read()

    def get_port(self):
        self._write(b"\x01\x11\x00\x00")
        response = self._read()
        if response == b"\x01\x11\x01\x00":
            port = 2
        elif response == b"\x01\x11\x00\x00":
            port = 1
        else:
            raise ValueError(f'invalid current port response {response}')
        logger.info(f'current port is [{port}]')
        return port

    def get_temp(self):
        self._write(b"\x01\x22\x00\x00")
        return self._read()

    def open(self):
        self.conn = serial.Serial(port=self.serial_port, baudrate=9600, parity=serial.PARITY_NONE, stopbits=serial.STOPBITS_ONE, bytesize=serial.EIGHTBITS)

    def close(self):
        self.conn.close()

    def __enter__(self):
        self.open()
        self.conn.flush()
        return self

    def __exit__(self, *args):
        self.conn.flush()
        self.close()


if __name__ == '__main__':
    with FFSW('/dev/serial/by-id/usb-FTDI_FT232R_USB_UART_AU03N2AV-if00-port0') as switch:
        print('flushing buffer')
        print(switch.conn.flush())
        print('setting port 1')
        print(switch.set_port(1))
        print('getting port after setting port 1')
        print(switch.get_port())
        print('setting port 2')
        print(switch.set_port(2))
        print('getting port after setting port 2')
        print(switch.get_port())
        print('getting temperature')
        print(switch.get_temp())
        print('getting module address')
        print(switch.get_module_address())


"""
Serial driver for Thorlabs FW102C filter wheel.

https://www.thorlabs.com/thorproduct.cfm?partnumber=FW102C#ad-image-0

Copyright (c) 2022, Jacob Feder
All rights reserved.
"""
import serial
import logging

logger = logging.getLogger(__name__)

class FW102C:
    def __init__(self, serial_port: str, channel_mapping=None, baud=115200, timeout: float = 10.0):
        """Args:
            serial_port: serial COM port (see pyserial docs)
            channel_mapping: dictionary mapping an identifier to a channel, which can be used instead of the channel number e.g.
            {'filter1': 1, 'filter2': 2}
            baud: 9600 or 115200
            timeout: read timeout
        """
        self.channel_mapping = channel_mapping
        self.serial_port = serial_port
        self.baud = baud
        self.timeout = timeout

    def __str__(self):
        return f'{self.serial_port} {self.idn}'

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        self.conn.close()

    def open(self):
        self.conn = serial.Serial(self.serial_port, baudrate=self.baud, timeout=self.timeout)
        # send a dummy command to clear any previous messages
        self.idn = self._query('*idn')
        logger.info(f'Connected to [{self}].')

    def _send_cmd(self, cmd):
        """Format and send a command to the device.
        
        Args:
            cmd: the command to send, e.g. '*idn?'
        
        Returns:
            The device response as a string (with the '>' stripped)
        """
        logger.debug(f'sending message [{cmd}]')
        cmd += '\r'
        self.conn.write(cmd.encode('ascii'))
        response = self.conn.read_until('> '.encode('ascii')).decode('ascii')
        if '> ' not in response: 
            raise RuntimeError('Timed out waiting for response')
        # strip preceding message and trailing '\r> '
        trimmed_response = response[len(cmd):-3]
        logger.debug(f'received response [{trimmed_response}]')
        return trimmed_response

    def _set(self, param, val):
        """Set a parameter value.
        
        Args:
            param: the parameter, e.g. 'pos' for the wheel position
            val: the value to set the parameter to
        """
        self._send_cmd(f'{param}={val}')

    def _query(self, param):
        """Get a parameter value.
        
        Args:
            param: the parameter, e.g. 'pos' for the wheel position
        """
        return self._send_cmd(f'{param}?')

    def version(self):
        """Return the version information"""
        return self._query('*idn')

    def position(self):
        """Return the wheel position."""
        pos_num = int(self._query('pos'))
        if self.channel_mapping is None:
            return pos_num
        else:
            inverse_channel_mapping = {v: k for k, v in self.channel_mapping.items()}
            return inverse_channel_mapping[pos_num]

    def set_position(self, pos):
        """Set the wheel position.

        Args:
            pos: new wheel position

        Raises:
            ValueError: invalid position
        """
        if self.channel_mapping is not None:
            pos = self.channel_mapping[pos]
        pcount = self.position_count()
        if pos in range(1, pcount+1):
            logger.info(f'setting position to [{pos}]')
            return self._set('pos', pos)
        else:
            raise ValueError(f'[{pos}] not in range [1, {pcount}]')

    def position_count(self):
        """Return the number of wheel positions."""
        return int(self._query('pcount'))

    def set_position_count(self, count):
        """Set the number of wheel positions.

        Args:
            count: number of wheel positions

        Raises:
            ValueError: invalid position count
        """
        if count != 6 and count != 12:
            raise ValueError(f'invalid position count [{count}]')
        logger.info(f'setting position count to [{count}]')
        return self._set('pcount', count)

    def trigger_mode(self):
        """Return the trigger mode."""
        return int(self._query('trig'))

    def set_trigger_mode(self, mode):
        """Set the trigger mode.

        Args:
            mode: 0 for input, 1 for output

        Raises:
            ValueError: invalid mode
        """
        if mode != 0 and mode != 1:
            raise ValueError(f'invalid trigger mode [{mode}]')
        logger.info(f'setting trigger mode to [{mode}]')
        return self._set('trig', mode)

    def speed_mode(self):
        """Return the speed mode."""
        return int(self._query('speed'))

    def set_speed_mode(self, mode):
        """Set the speed mode.

        Args:
            mode: 0 for slow, 1 for fast

        Raises:
            ValueError: invalid mode
        """
        if mode != 0 and mode != 1:
            raise ValueError(f'invalid speed mode [{mode}]')
        logger.info(f'setting speed mode to [{mode}]')
        return self._set('speed', mode)

    def sensor_mode(self):
        """Return the sensor mode."""
        return int(self._query('sensors'))

    def set_sensor_mode(self, mode):
        """Set the sensor mode.

        Args:
            mode: 0 to turn off sensors when wheel is idle to eliminate stray light, 1 to keep sensors active 

        Raises:
            ValueError: invalid mode
        """
        if mode != 0 and mode != 1:
            raise ValueError(f'invalid speed mode [{mode}]')
        logger.info(f'setting sensor mode to [{mode}]')
        return self._set('sensors', mode)

    def baud(self):
        """Return the baud rate."""
        return int(self._query('baud'))

    def set_baud(self, baud):
        """Set the baud rate.

        Args:
            baud: 0 for 9600, 1 for 115200

        Raises:
            ValueError: invalid baud
        """
        if baud != 0 and baud != 1:
            raise ValueError(f'invalid baud [{baud}]')
        logger.info(f'setting baud to [{baud}]')
        return self._set('baud', baud)

    def save(self):
        """Save the current settings as default"""
        self._send_cmd('save')



"""
GUI for the thorlabs FW102C filter wheel.

https://www.thorlabs.com/thorproduct.cfm?partnumber=FW102C#ad-image-0

Copyright (c) 2022, Jacob Feder
All rights reserved.
"""

from pyqtgraph.Qt import QtWidgets
from pyqtgraph import SpinBox

class FW102CWidget(QtWidgets.QWidget):
    """Qt widget for controlling the Thorlabs FW102C filter wheel."""

    def __init__(self, filter_wheel_driver, filter_wheel_channel_mapping):
        """
        Args:
            filter_wheel_driver: The FW102C driver.
            filter_wheel_channel_mapping: TODO.
        """
        super().__init__()

        self.filter_wheel = filter_wheel_driver

        # top level layout
        layout = QtWidgets.QVBoxLayout()

        wheel_position_layout = QtWidgets.QHBoxLayout()
        position_label = QtWidgets.QLabel('Position')
        position_label.setSizePolicy(
            QtWidgets.QSizePolicy(
                QtWidgets.QSizePolicy.Policy.Fixed,
                QtWidgets.QSizePolicy.Policy.Fixed,
            )
        )
        wheel_position_layout.addWidget(position_label)
        filter_wheel_combobox = QtWidgets.QComboBox()
        for m in filter_wheel_channel_mapping:
            filter_wheel_combobox.addItem(m)
        filter_wheel_combobox.setCurrentText(self.filter_wheel.position())
        filter_wheel_combobox.currentTextChanged.connect(lambda text: self.filter_wheel.set_position(text))
        wheel_position_layout.addWidget(filter_wheel_combobox)

        layout.addLayout(wheel_position_layout)
        layout.addStretch()

        self.setLayout(layout)



"""
Qt GUI for Swabian/LABS electronics DLnsec laser.

Copyright (c) 2022, Benjamin Soloway, Jacob Feder
All rights reserved.
"""
import logging

from pyqtgraph.Qt import QtWidgets
from pyqtgraph import SpinBox

logger = logging.getLogger(__name__)

class DLnsecWidget(QtWidgets.QWidget):
    """Qt widget for controlling DLnsec lasers."""
    def __init__(self, laser_driver):
        """
        Args:
            laser_driver: The dlnsec driver.
        """
        super().__init__()

        self.laser = laser_driver

        # top level layout
        layout = QtWidgets.QGridLayout()
        layout_row = 0

        # button to turn the laser off
        self.off_button = QtWidgets.QPushButton('Off')
        self.off_button.clicked.connect(lambda: self.laser.off())
        layout.addWidget(self.off_button, layout_row, 0)

        # button to turn the laser on
        self.on_button = QtWidgets.QPushButton('On')
        self.on_button.clicked.connect(lambda: self.laser.on())
        layout.addWidget(self.on_button, layout_row, 1)

        # button to put the laser in CW mode
        self.on_button = QtWidgets.QPushButton('CW')
        self.on_button.clicked.connect(lambda: self.laser.cw_mode())
        layout.addWidget(self.on_button, layout_row, 2)

        # button to reboot the laser
        self.on_button = QtWidgets.QPushButton('Reboot')
        self.on_button.clicked.connect(lambda: self.laser.reboot())
        layout.addWidget(self.on_button, layout_row, 3)
        layout_row += 1

        # power spinbox
        layout.addWidget(QtWidgets.QLabel('Power (fraction from 0-1)'), layout_row, 0)
        self.power_spinbox = SpinBox(value=0, siPrefix=False, bounds=(0, 1), dec=True, minStep=0.01)
        def set_power(spinbox):
            self.laser.power(int(spinbox.value()*100))
        self.power_spinbox.sigValueChanged.connect(set_power)
        self.power_spinbox.setValue(value=0)
        layout.addWidget(self.power_spinbox, layout_row, 1)
        layout_row += 1

        # # dropdown menu to pick laser mode
        # layout.addWidget(QtWidgets.QLabel('Mode'), layout_row, 0)
        # self.mode_dropdown = QtWidgets.QComboBox()
        # #self.mode_dropdown.addItem('Digital Modulation') # index 0
        # self.mode_dropdown.addItem('Constant Power') # index 1
        # # default to modulation mode
        # self.mode_dropdown.setCurrentIndex(0)
        # self.mode_dropdown.currentIndexChanged.connect(self.change_mode)
        # layout.addWidget(self.mode_dropdown, layout_row, 1)
        # layout_row += 1

        # # state label
        # self.state_label = QtWidgets.QLabel('')
        # self.update_state()
        # layout.addWidget(self.state_label, layout_row, 1)
        # # get state button
        # self.get_state_button = QtWidgets.QPushButton('Get State')
        # self.get_state_button.clicked.connect(self.update_state)
        # layout.addWidget(self.get_state_button, layout_row, 0)
        # layout_row += 1

        # take up any additional space in the final column with padding
        layout.setColumnStretch(2, 1)
        # take up any additional space in the final row with padding
        layout.setRowStretch(layout_row, 1)

        self.setLayout(layout)

    # def update_state(self):
    #     """Query the laser for the current state then update the state text box."""
    #     self.state_label.setText(self.laser.get_state()[4:])

    # def change_mode(self, idx):
    #     """Change the laser output mode."""
    #     # constant power
    #     # digital modulation
    #     if idx == 0:
    #         self.laser.modulation_mode()
    #     elif idx == 1:
    #         self.laser.constant_power()
    #     else:
    #         raise ValueError('Invalid mode selection')
    #     self.update_state()



"""
Ximea camera driver.

Copyright (c) 2022, Jacob Feder
All rights reserved.
"""
import ximea.xiapi
import numpy as np

class XIMEA:
    """XIMEA camera wrapper driver."""
    def __init__(self, serial_number=None):
        self.serial_number = serial_number

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, *args):
        self.disconnect()

    def connect(self):
        if self.serial_number is not None:
            self.driver = ximea.xiapi.open_device_by_SN(serial_number)
        else:
            # create instance for first connected camera
            self.driver = ximea.xiapi.Camera()

        self.driver.open_device()
        # default settings
        # exposure (us)
        self.driver.set_exposure(100000)

    def disconnect(self):
        self.driver.close_device()

    def start_acquisition(self):
        self.driver.start_acquisition()

    def stop_acquisition(self):
        self.driver.stop_acquisition()

    def get_image(self):
        img = ximea.xiapi.Image()
        self.driver.get_image(img)
        # get raw data from camera
        # for Python3.x function returns bytes
        data_raw = img.get_image_data_raw()

        # transform data to numpy array
        data = np.array(list(data_raw), dtype=np.uint8).reshape(img.width, img.height)

        return data



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['nspyre_drivers',
 'nspyre_drivers.acton.sp2150i',
 'nspyre_drivers.agiltron',
 'nspyre_drivers.agiltron.ffsw',
 'nspyre_drivers.arduino',
 'nspyre_drivers.arduino.arduino_gpio',
 'nspyre_drivers.beaglebone',
 'nspyre_drivers.beaglebone.beaglebone_gpio',
 'nspyre_drivers.beaglebone.beaglebone_gpio.beaglebone_server',
 'nspyre_drivers.cobolt',
 'nspyre_drivers.cv2_cam',
 'nspyre_drivers.dli_pdu',
 'nspyre_drivers.dlnsec',
 'nspyre_drivers.luminox',
 'nspyre_drivers.ni',
 'nspyre_drivers.oceanoptics',
 'nspyre_drivers.oceanoptics.hr2000es',
 'nspyre_drivers.oceanoptics.hr2000es.oceandirect.sample_pack.Python.Python '
 'VVV',
 'nspyre_drivers.pi',
 'nspyre_drivers.pi.pi710',
 'nspyre_drivers.rigol',
 'nspyre_drivers.rigol.dp832',
 'nspyre_drivers.rohde_schwarz',
 'nspyre_drivers.rohde_schwarz.hmp4040',
 'nspyre_drivers.thorlabs',
 'nspyre_drivers.thorlabs.cld1010',
 'nspyre_drivers.thorlabs.fw102c',
 'nspyre_drivers.thorlabs.pm100d',
 'nspyre_drivers.ximea',
 'nspyre_drivers.zaber']

package_data = \
{'': ['*'],
 'nspyre_drivers': ['rfsoc/*'],
 'nspyre_drivers.arduino.arduino_gpio': ['pin_server/*'],
 'nspyre_drivers.oceanoptics.hr2000es': ['oceandirect/*',
                                         'oceandirect/sample_pack/*',
                                         'oceandirect/sample_pack/C/*',
                                         'oceandirect/sample_pack/CPlusPlus/*',
                                         'oceandirect/sample_pack/CSharp/*',
                                         'oceandirect/sample_pack/LabVIEW/LabVIEW '
                                         '2010/*',
                                         'oceandirect/sample_pack/LabVIEW/LabVIEW '
                                         '2020/*',
                                         'oceandirect/sample_pack/MATLAB/MATLAB '
                                         'VVV/*']}

install_requires = \
['numpy', 'pyserial', 'pyusb', 'pyvisa', 'pyvisa-py']

extras_require = \
{'beaglebone': ['requests'],
 'dli_pdu': ['beautifulsoup4', 'dlipower'],
 'nidaq': ['nidaqmx'],
 'oceanoptics': ['seabreeze'],
 'ximea': ['ximea-py'],
 'zaber': ['zaber-motion']}

setup_kwargs = {
    'name': 'nspyre-drivers',
    'version': '0.1.6',
    'description': 'A set of Python drivers for lab instrumentation.',
    'long_description': "# Drivers\nThis is a set of Python drivers for lab instrumentation. These drivers are \nassociated with [nspyre](https://nspyre.readthedocs.io/en/latest/), but are \nalso suitable for general usage. Unless otherwise specified, drivers are \nprovided under the terms of the MIT license.\n\n## Installation\n\n```bash\npip install nspyre-drivers\n```\n\nCertain drivers require extra dependencies. Those dependencies can be installed \nby specifying a tag during the install. E.g. to install the DLI pdu driver \ndependencies, use:\n\n```bash\npip install nspyre-drivers[dli_pdu]\n```\n\nA full listing of the tags is below.\n\n```\nbeaglebone\ndli_pdu\noceanoptics\nximea\nzaber\n```\n\nFor some USB drivers on Linux, you need to grant user access to the drivers in \norder for VISA to detect them:\nYou should find the udev rules file in the same folder as the driver, then, e.g.:\n\n```bash\nsudo cp src/drivers/thorlabs/cld1010/99-thorlabs-cld1010.rules /etc/udev/rules.d/\n````\n\nCreate a user group for the usb device access:\n\n```bash\nsudo groupadd usbtmc\n```\n\nAdd any relevant users to the group:\n\n```bash\nusermod -aG usbtmc <myuser>\n```\n\nReboot for the changes to take effect.\n\n## Usage\n\nAfter installation, you can use drivers with, e.g.:\n\n```python\nfrom nspyre_drivers.rohde_schwarz.hmp4040.hmp4040 import HMP4040\nwith HMP4040('TCPIP::192.168.0.10::5025::SOCKET') as power_supply:\n\t# set the power supply channel 1 to 5V\n\tpower_supply.set_voltage(1, 5.0)\n```\n\nSee the source for all of the available drivers and the import paths.\n\n## Other Drivers\n\nIn order to minimize reinventing the wheel, below is a list of other sources of \npython instrument drivers. Please contribute if you find other useful sources!\n\n[pycobolt](https://github.com/cobolt-lasers/pycobolt)\n\n## Contributing\n\nIf you want to contribute driver code, please submit it as a \n[pull request](https://nspyre.readthedocs.io/en/latest/contributing.html#forking-pull-requests). This project uses \n[poetry](https://python-poetry.org/). If your driver requires specific \ndependencies beyond those currently in use in the project, you should include \nthem in the pyproject.toml file as extras. See the poetry documentation for \nspecifics on how to declare these dependencies.\n",
    'author': 'Jacob Feder',
    'author_email': 'jacobsfeder@gmail.com',
    'maintainer': 'Jacob Feder',
    'maintainer_email': 'jacobsfeder@gmail.com',
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Driver for the Thorlabs CLD1010LP.

Programming manual: 
https://www.thorlabs.com/drawings/84b68761fa2c704c-5D5DD69C-9E91-8A48-043C41D5A362BD66/CLD1010LP-ProgrammersReferenceManual.pdf

Copyright (c) 2022, Jacob Feder, Ben Soloway
All rights reserved.
"""
import logging

from pyvisa import ResourceManager

logger = logging.getLogger(__name__)

class CLD1010:
    def __init__(self, address):
        """
        Args:
            address: PyVISA resource path.
        """
        self.rm = ResourceManager('@py')
        self.address = address

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *args):
        self.off()
        self.close()

    def __str__(self):
        return f'{self.address} {self.idn}'

    def open(self):
        try:
            self.laser = self.rm.open_resource(self.address)
        except Exception as err:
            raise ConnectionError(f'Failed connecting to CLD1010 @ [{self.address}]') from err
        # 1 second timeout
        self.laser.timeout = 1000
        self.idn = self.laser.query('*IDN?')
        logger.info(f'Connected to CLD1010 [{self}].')
        return self

    def close(self):
        self.laser.close()

    def idn(self):
        return self.laser.query('*IDN?')

    def get_ld_state(self):
        return int(self.laser.query('OUTP1:STAT?'))

    def set_ld_state(self, value):
        self.laser.write(f'OUTP1:STAT {value}')

    def max_current(self):
        return float(self.laser.query('SOUR:CURR:LIM:AMPL?'))

    def meas_current(self):
        return float(self.laser.query('MEAS:CURR?'))

    def get_current_setpoint(self):
        return float(self.laser.query('SOUR:CURR?'))

    def set_current_setpoint(self, value):
        max_current = self.max_current()
        if value <= max_current:
            self.laser.write(f'SOUR:CURR {value:.5f}')
        else:
            raise ValueError(f'Current setpoint: [{value}] is larger than max current [{max_current}])')

    def get_tec_state(self):
        return self.laser.query('OUTP2:STAT?')

    def set_tec_state(self, value):
        self.laser.write(f'OUTP2:STAT {value}')

    def temperature(self):
        return self.laser.query('MEAS:TEMP?')

    def on(self):
        if self.get_tec_state():
            self.set_ld_state(1)
        else:
            return("error: temperature controller not on")

    def get_modulation_state(self):
        value = int(self.laser.query('SOUR:AM:STAT?'))
        if value == 0:
            return 'Off'
        elif value == 1:
            return 'On'
        else:
            raise ValueError(f'invalid modulation state {value}')

    def set_modulation_state(self, value):
        if value == 'Off':
            val = 0
        elif value == 'On':
            val = 1
        else:
            raise ValueError(f'invalid modulation state {value}')
        self.laser.write(f'SOUR:AM:STAT {val}')

    def off(self):
        self.set_ld_state(0)
